from typing import (
    Tuple,
)

from eth_utils import (
    ValidationError,
    big_endian_to_int,
)
from eth_utils.toolz import (
    curry,
    pipe,
)
from py_ecc import (
    optimized_bn128 as bn128,
)

from eth import (
    constants,
)
from eth._utils.bn128 import (
    FQP_point_to_FQ2_point,
    validate_point,
)
from eth._utils.padding import (
    pad32,
)
from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    VMError,
)
from eth.typing import (
    BytesOrView,
)

ZERO = bn128.Z2
EXPONENT = bn128.FQ12.one()


@curry
def ecpairing(
    computation: ComputationAPI,
    gas_cost_base: int = constants.GAS_ECPAIRING_BASE,
    gas_cost_per_point: int = constants.GAS_ECPAIRING_PER_POINT,
) -> ComputationAPI:
    if len(computation.msg.data) % 192:
        # data length must be an exact multiple of 192
        raise VMError("Invalid ECPAIRING parameters")

    num_points = len(computation.msg.data) // 192
    gas_fee = gas_cost_base + num_points * gas_cost_per_point

    computation.consume_gas(gas_fee, reason="ECPAIRING Precompile")

    try:
        result = _ecpairing(computation.msg.data)
    except ValidationError:
        raise VMError("Invalid ECPAIRING parameters")

    if result is True:
        computation.output = pad32(b"\x01")
    elif result is False:
        computation.output = pad32(b"\x00")
    else:
        raise Exception("Invariant: unreachable code path")
    return computation


def _ecpairing(data: BytesOrView) -> bool:
    exponent = bn128.FQ12.one()

    processing_pipeline = (
        _process_point(data[start_idx : start_idx + 192])
        for start_idx in range(0, len(data), 192)
    )
    exponent = pipe(bn128.FQ12.one(), *processing_pipeline)

    result = bn128.final_exponentiate(exponent) == bn128.FQ12.one()
    return result


@curry
def _process_point(data_buffer: bytes, exponent: int) -> bn128.FQP:
    x1, y1, x2_i, x2_r, y2_i, y2_r = _extract_point(data_buffer)
    p1 = validate_point(x1, y1)

    for v in (x2_i, x2_r, y2_i, y2_r):
        if v >= bn128.field_modulus:
            raise ValidationError("value greater than field modulus")

    fq2_x = bn128.FQ2([x2_r, x2_i])
    fq2_y = bn128.FQ2([y2_r, y2_i])

    p2 = ZERO
    if (fq2_x, fq2_y) != (bn128.FQ2.zero(), bn128.FQ2.zero()):
        p2 = (fq2_x, fq2_y, bn128.FQ2.one())
        if not bn128.is_on_curve(p2, bn128.b2):
            raise ValidationError("point is not on curve")

    if bn128.multiply(p2, bn128.curve_order)[-1] != bn128.FQ2.zero():
        raise ValidationError("TODO: what case is this?????")

    return exponent * bn128.pairing(
        FQP_point_to_FQ2_point(p2), p1, final_exponentiate=False
    )


def _extract_point(data_slice: bytes) -> Tuple[int, int, int, int, int, int]:
    x1_bytes = data_slice[:32]
    y1_bytes = data_slice[32:64]
    x2_i_bytes = data_slice[64:96]
    x2_r_bytes = data_slice[96:128]
    y2_i_bytes = data_slice[128:160]
    y2_r_bytes = data_slice[160:192]

    x1 = big_endian_to_int(x1_bytes)
    y1 = big_endian_to_int(y1_bytes)
    x2_i = big_endian_to_int(x2_i_bytes)
    x2_r = big_endian_to_int(x2_r_bytes)
    y2_i = big_endian_to_int(y2_i_bytes)
    y2_r = big_endian_to_int(y2_r_bytes)

    return x1, y1, x2_i, x2_r, y2_i, y2_r



from eth_typing import (
    BlockNumber,
    Hash32,
)

from eth.abc import (
    SchemaAPI,
)


class SchemaV1(SchemaAPI):
    @staticmethod
    def make_canonical_head_hash_lookup_key() -> bytes:
        return b"v1:canonical_head_hash"

    @staticmethod
    def make_block_number_to_hash_lookup_key(block_number: BlockNumber) -> bytes:
        number_to_hash_key = b"block-number-to-hash:%d" % block_number
        return number_to_hash_key

    @staticmethod
    def make_block_hash_to_score_lookup_key(block_hash: Hash32) -> bytes:
        return b"block-hash-to-score:%s" % block_hash

    @staticmethod
    def make_header_chain_gaps_lookup_key() -> bytes:
        return b"v1:header_chain_gaps"

    @staticmethod
    def make_chain_gaps_lookup_key() -> bytes:
        return b"v1:chain_gaps"

    @staticmethod
    def make_checkpoint_headers_key() -> bytes:
        """
        Checkpoint header hashes stored as concatenated 32 byte values
        """
        return b"v1:checkpoint-header-hashes-list"

    @staticmethod
    def make_transaction_hash_to_block_lookup_key(transaction_hash: Hash32) -> bytes:
        return b"transaction-hash-to-block:%s" % transaction_hash

    @staticmethod
    def make_withdrawal_hash_to_block_lookup_key(withdrawal_hash: Hash32) -> bytes:
        return b"withdrawal-hash-to-block:%s" % withdrawal_hash



import contextlib
import itertools
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from cached_property import (
    cached_property,
)
from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    encode_hex,
)
import rlp
from web3 import Web3

from eth._utils.datatypes import (
    Configurable,
)
from eth._utils.db import (
    get_block_header_by_hash,
    get_parent_header,
)
from eth.abc import (
    AtomicDatabaseAPI,
    BlockAndMetaWitness,
    BlockAPI,
    BlockHeaderAPI,
    ChainContextAPI,
    ChainDatabaseAPI,
    ComputationAPI,
    ConsensusAPI,
    ConsensusContextAPI,
    ExecutionContextAPI,
    ReceiptAPI,
    ReceiptBuilderAPI,
    SignedTransactionAPI,
    StateAPI,
    TransactionBuilderAPI,
    UnsignedTransactionAPI,
    VirtualMachineAPI,
    WithdrawalAPI,
)
from eth.consensus.pow import (
    PowConsensus,
)
from eth.constants import (
    GENESIS_PARENT_HASH,
    MAX_PREV_HEADER_DEPTH,
    MAX_UNCLES,
)
from eth.db.trie import (
    make_trie_root_and_nodes,
)
from eth.exceptions import (
    HeaderNotFound,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.rlp.sedes import (
    uint32,
)
from eth.validation import (
    validate_gas_limit,
    validate_length_lte,
)
from eth.vm.execution_context import (
    ExecutionContext,
)
from eth.vm.interrupt import (
    EVMMissingData,
)
from eth.vm.message import (
    Message,
)
from web3.types import (
    BlockIdentifier,
)

if TYPE_CHECKING:
    from eth.typing import (  # noqa: F401
        Block,
    )


class VM(Configurable, VirtualMachineAPI):
    block_class: Type[BlockAPI] = None
    consensus_class: Type[ConsensusAPI] = PowConsensus
    extra_data_max_bytes: ClassVar[int] = 32
    fork: str = None
    chaindb: ChainDatabaseAPI = None
    _state_class: Type[StateAPI] = None

    _state = None
    _block = None
    _web3: Web3 = None
    _block_identifier: BlockIdentifier = None

    cls_logger = logging.getLogger("eth.vm.base.VM")

    def __init__(
        self,
        header: BlockHeaderAPI,
        chaindb: ChainDatabaseAPI,
        chain_context: ChainContextAPI,
        consensus_context: ConsensusContextAPI,
        web3: Web3,
        block_identifier: BlockIdentifier = None,
    ) -> None:
        self.chaindb = chaindb
        self.chain_context = chain_context
        self.consensus_context = consensus_context
        self._web3 = web3
        self._block_identifier = block_identifier
        self._initial_header = header

    def get_header(self) -> BlockHeaderAPI:
        if self._block is None:
            return self._initial_header
        else:
            return self._block.header

    def get_block(self) -> BlockAPI:
        if self._block is None:
            block_class = self.get_block_class()
            self._block = block_class.from_header(
                header=self._initial_header, chaindb=self.chaindb
            )
        return self._block

    @property
    def state(self) -> StateAPI:
        if self._state is None:
            self._state = self.build_state(
                self._web3,
                self._block_identifier,
                self.get_header(),
                self.chain_context,
                self.previous_hashes,
            )
        return self._state

    @classmethod
    def build_state(
        cls,
        web3: Web3,
        block_identifier: BlockIdentifier,
        header: BlockHeaderAPI,
        chain_context: ChainContextAPI,
        previous_hashes: Iterable[Hash32] = (),
    ) -> StateAPI:
        execution_context = cls.create_execution_context(
            header, previous_hashes, chain_context
        )
        return cls.get_state_class()(web3, block_identifier, execution_context)

    @cached_property
    def _consensus(self) -> ConsensusAPI:
        return self.consensus_class(self.consensus_context)

    #
    # Logging
    #
    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(f"eth.vm.base.VM.{self.__class__.__name__}")

    #
    # Execution
    #
    def apply_transaction(
        self, header: BlockHeaderAPI, transaction: SignedTransactionAPI
    ) -> Tuple[ReceiptAPI, ComputationAPI]:
        self.validate_transaction_against_header(header, transaction)

        # Mark current state as un-revertable, since new transaction is starting...
        self.state.lock_changes()

        computation = self.state.apply_transaction(transaction)
        receipt = self.make_receipt(header, transaction, computation, self.state)
        self.validate_receipt(receipt)

        return receipt, computation

    @classmethod
    def create_execution_context(
        cls,
        header: BlockHeaderAPI,
        prev_hashes: Iterable[Hash32],
        chain_context: ChainContextAPI,
    ) -> ExecutionContextAPI:
        fee_recipient = cls.consensus_class.get_fee_recipient(header)
        try:
            base_fee = header.base_fee_per_gas
        except AttributeError:
            return ExecutionContext(
                coinbase=fee_recipient,
                timestamp=header.timestamp,
                block_number=header.block_number,
                difficulty=header.difficulty,
                mix_hash=header.mix_hash,
                gas_limit=header.gas_limit,
                prev_hashes=prev_hashes,
                chain_id=chain_context.chain_id,
            )
        else:
            return ExecutionContext(
                coinbase=fee_recipient,
                timestamp=header.timestamp,
                block_number=header.block_number,
                difficulty=header.difficulty,
                mix_hash=header.mix_hash,
                gas_limit=header.gas_limit,
                prev_hashes=prev_hashes,
                chain_id=chain_context.chain_id,
                base_fee_per_gas=base_fee,
            )

    def execute_bytecode(
        self,
        origin: Address,
        gas_price: int,
        gas: int,
        to: Address,
        sender: Address,
        value: int,
        data: bytes,
        code: bytes,
        code_address: Address = None,
    ) -> ComputationAPI:
        if origin is None:
            origin = sender

        # Construct a message
        message = Message(
            gas=gas,
            to=to,
            sender=sender,
            value=value,
            data=data,
            code=code,
            code_address=code_address,
        )

        # Construction a tx context
        transaction_context = self.state.get_transaction_context_class()(
            gas_price=gas_price,
            origin=origin,
        )

        # Execute it in the VM
        return self.state.computation_class.apply_computation(
            self.state,
            message,
            transaction_context,
        )

    def apply_all_transactions(
        self, transactions: Sequence[SignedTransactionAPI], base_header: BlockHeaderAPI
    ) -> Tuple[BlockHeaderAPI, Tuple[ReceiptAPI, ...], Tuple[ComputationAPI, ...]]:
        vm_header = self.get_header()
        if base_header.block_number != vm_header.block_number:
            raise ValidationError(
                f"This VM instance must only work on block #{self.get_header().block_number}, "  # noqa: E501
                f"but the target header has block #{base_header.block_number}"
            )

        receipts = []
        computations = []
        previous_header = base_header
        result_header = base_header

        for transaction_index, transaction in enumerate(transactions):
            snapshot = self.state.snapshot()
            try:
                receipt, computation = self.apply_transaction(
                    previous_header,
                    transaction,
                )
            except EVMMissingData:
                self.state.revert(snapshot)
                raise

            result_header = self.add_receipt_to_header(previous_header, receipt)
            previous_header = result_header
            receipts.append(receipt)
            computations.append(computation)

            self.transaction_applied_hook(
                transaction_index,
                transactions,
                vm_header,
                result_header,
                computation,
                receipt,
            )

        receipts_tuple = tuple(receipts)
        computations_tuple = tuple(computations)

        return result_header, receipts_tuple, computations_tuple

    #
    # Withdrawals
    #
    def apply_withdrawal(
        self,
        withdrawal: WithdrawalAPI,
    ) -> None:
        self.state.apply_withdrawal(withdrawal)

    def apply_all_withdrawals(self, withdrawals: Sequence[WithdrawalAPI]) -> None:
        touched_addresses: List[Address] = []

        for withdrawal in withdrawals:
            # validate withdrawal fields
            withdrawal.validate()

            self.apply_withdrawal(withdrawal)

            # collect all touched addresses
            if withdrawal.address not in touched_addresses:
                touched_addresses.append(withdrawal.address)

        for address in touched_addresses:
            # if account is empty after applying all withdrawals, delete it
            if self.state.account_is_empty(address):
                self.state.delete_account(address)

    #
    # Importing blocks
    #
    def import_block(self, block: BlockAPI) -> BlockAndMetaWitness:
        if self.get_block().number != block.number:
            raise ValidationError(
                f"This VM can only import blocks at number #{self.get_block().number},"
                f" the attempted block was #{block.number}"
            )

        header_params = {
            "coinbase": block.header.coinbase,
            "difficulty": block.header.difficulty,
            "gas_limit": block.header.gas_limit,
            "timestamp": block.header.timestamp,
            "extra_data": block.header.extra_data,
            "mix_hash": block.header.mix_hash,
            "nonce": block.header.nonce,
            "uncles_hash": keccak(rlp.encode(block.uncles)),
        }

        block_params = {
            "header": self.configure_header(**header_params),
            "uncles": block.uncles,
        }

        if hasattr(block, "withdrawals"):
            # post-shanghai blocks
            block_params["withdrawals"] = block.withdrawals

        self._block = self.get_block().copy(**block_params)

        execution_context = self.create_execution_context(
            block.header, self.previous_hashes, self.chain_context
        )

        # Zero out the gas_used before applying transactions. Each applied transaction
        # will increase the gas used in the final new_header.
        header = self.get_header().copy(gas_used=0)

        # we need to re-initialize the `state` to update the execution context.
        self._state = self.get_state_class()(
            self.chaindb.db, execution_context, header.state_root
        )

        # run all of the transactions.
        new_header, receipts, _ = self.apply_all_transactions(
            block.transactions, header
        )

        withdrawals = block.withdrawals if hasattr(block, "withdrawals") else None

        if withdrawals:
            # post-shanghai blocks
            self.apply_all_withdrawals(block.withdrawals)

        filled_block = self.set_block_transactions_and_withdrawals(
            self.get_block(),
            new_header,
            block.transactions,
            receipts,
            withdrawals=withdrawals,
        )

        return self.mine_block(filled_block)

    def mine_block(
        self, block: BlockAPI, *args: Any, **kwargs: Any
    ) -> BlockAndMetaWitness:
        packed_block = self.pack_block(block, *args, **kwargs)
        block_result = self.finalize_block(packed_block)

        # Perform validation
        self.validate_block(block_result.block)

        return block_result

    def set_block_transactions_and_withdrawals(
        self,
        base_block: BlockAPI,
        new_header: BlockHeaderAPI,
        transactions: Sequence[SignedTransactionAPI],
        receipts: Sequence[ReceiptAPI],
        withdrawals: Sequence[WithdrawalAPI] = None,
    ) -> BlockAPI:
        tx_root_hash, tx_kv_nodes = make_trie_root_and_nodes(transactions)
        self.chaindb.persist_trie_data_dict(tx_kv_nodes)

        receipt_root_hash, receipt_kv_nodes = make_trie_root_and_nodes(receipts)
        self.chaindb.persist_trie_data_dict(receipt_kv_nodes)

        block_fields: "Block" = {"transactions": transactions}
        block_header_fields = {
            "transaction_root": tx_root_hash,
            "receipt_root": receipt_root_hash,
        }

        if withdrawals:
            withdrawals_root_hash, withdrawal_kv_nodes = make_trie_root_and_nodes(
                withdrawals,
            )
            self.chaindb.persist_trie_data_dict(withdrawal_kv_nodes)

            block_fields["withdrawals"] = withdrawals
            block_header_fields["withdrawals_root"] = withdrawals_root_hash

        block_fields["header"] = new_header.copy(**block_header_fields)

        return base_block.copy(**block_fields)

    #
    # Finalization
    #
    def _assign_block_rewards(self, block: BlockAPI) -> None:
        block_reward = self.get_block_reward() + (
            len(block.uncles) * self.get_nephew_reward()
        )

        # EIP-161:
        # Even if block reward is zero, the coinbase is still touched here. This was
        # not likely to ever happen in PoW, except maybe in some very niche cases, but
        # happens now in PoS. In these cases, the coinbase may end up zeroed after the
        # computation and thus should be marked for deletion since it was touched.
        self.state.delta_balance(block.header.coinbase, block_reward)
        self.logger.debug(
            "BLOCK REWARD: %s -> %s",
            block_reward,
            encode_hex(block.header.coinbase),
        )

        for uncle in block.uncles:
            uncle_reward = self.get_uncle_reward(block.number, uncle)
            self.logger.debug(
                "UNCLE REWARD REWARD: %s -> %s",
                uncle_reward,
                encode_hex(uncle.coinbase),
            )
            self.state.delta_balance(uncle.coinbase, uncle_reward)

    def finalize_block(self, block: BlockAPI) -> BlockAndMetaWitness:
        if block.number > 0:
            snapshot = self.state.snapshot()
            try:
                self._assign_block_rewards(block)
            except EVMMissingData:
                self.state.revert(snapshot)
                raise
            else:
                self.state.commit(snapshot)

        # We need to call `persist` here since the state db batches
        # all writes until we tell it to write to the underlying db
        meta_witness = self.state.persist()

        final_block = block.copy(
            header=block.header.copy(state_root=self.state.state_root)
        )

        self.logger.debug(
            "%s reads %d unique node hashes, %d addresses, %d bytecodes, and %d storage slots",  # noqa: E501
            final_block,
            len(meta_witness.hashes),
            len(meta_witness.accounts_queried),
            len(meta_witness.account_bytecodes_queried),
            meta_witness.total_slots_queried,
        )

        return BlockAndMetaWitness(final_block, meta_witness)

    def pack_block(self, block: BlockAPI, *args: Any, **kwargs: Any) -> BlockAPI:
        if "uncles" in kwargs:
            uncles = kwargs.pop("uncles")
            kwargs.setdefault("uncles_hash", keccak(rlp.encode(uncles)))
        else:
            uncles = block.uncles

        provided_fields = set(kwargs.keys())
        known_fields = set(BlockHeader._meta.field_names)
        unknown_fields = provided_fields.difference(known_fields)

        if unknown_fields:
            raise AttributeError(
                f"Unable to set the field(s) {', '.join(known_fields)} "
                "on the `BlockHeader` class. "
                f"Received the following unexpected fields: {', '.join(unknown_fields)}."  # noqa: E501
            )

        header: BlockHeaderAPI = block.header.copy(**kwargs)

        packed_block = block.copy(uncles=uncles, header=header)

        return packed_block

    #
    # Blocks
    #
    @classmethod
    def generate_block_from_parent_header_and_coinbase(
        cls, parent_header: BlockHeaderAPI, coinbase: Address
    ) -> BlockAPI:
        block_header = cls.create_header_from_parent(parent_header, coinbase=coinbase)
        block = cls.get_block_class()(
            block_header,
            transactions=[],
            uncles=[],
        )
        return block

    @classmethod
    def create_genesis_header(cls, **genesis_params: Any) -> BlockHeaderAPI:
        # Create genesis header by setting the parent to None
        return cls.create_header_from_parent(None, **genesis_params)

    @classmethod
    def get_block_class(cls) -> Type[BlockAPI]:
        if cls.block_class is None:
            raise AttributeError("No `block_class` has been set for this VM")
        else:
            return cls.block_class

    @classmethod
    def get_prev_hashes(
        cls, last_block_hash: Hash32, chaindb: ChainDatabaseAPI
    ) -> Optional[Iterable[Hash32]]:
        if last_block_hash == GENESIS_PARENT_HASH:
            return

        # stupid - but we need to import this here to avoid circular imports
        # and I don't want to refactor the whole module right now.
        from eth.db.header import build_block_header


        block_header = build_block_header(chaindb.w3, last_block_hash)

        for _ in range(MAX_PREV_HEADER_DEPTH):
            yield last_block_hash
            try:
                last_block_hash = block_header.parent_hash
                block_header = build_block_header(chaindb.w3, block_header.parent_hash)
            except (IndexError, HeaderNotFound):
                break

    @property
    def previous_hashes(self) -> Optional[Iterable[Hash32]]:
        return self.get_prev_hashes(self.get_header().parent_hash, self.chaindb)

    #
    # Transactions
    #
    def create_transaction(self, *args: Any, **kwargs: Any) -> SignedTransactionAPI:
        return self.get_transaction_builder().new_transaction(*args, **kwargs)

    @classmethod
    def create_unsigned_transaction(
        cls,
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
    ) -> UnsignedTransactionAPI:
        return cls.get_transaction_builder().create_unsigned_transaction(
            nonce=nonce, gas_price=gas_price, gas=gas, to=to, value=value, data=data
        )

    @classmethod
    def get_transaction_builder(cls) -> Type[TransactionBuilderAPI]:
        return cls.get_block_class().get_transaction_builder()

    @classmethod
    def get_receipt_builder(cls) -> Type[ReceiptBuilderAPI]:
        return cls.get_block_class().get_receipt_builder()

    #
    # Validate
    #
    @classmethod
    def validate_receipt(cls, receipt: ReceiptAPI) -> None:
        already_checked: Set[Union[Address, int]] = set()

        for log_idx, log in enumerate(receipt.logs):
            if log.address in already_checked:
                continue
            elif log.address not in receipt.bloom_filter:
                raise ValidationError(
                    f"The address from the log entry at position {log_idx} is not "
                    "present in the provided bloom filter."
                )
            already_checked.add(log.address)

        for log_idx, log in enumerate(receipt.logs):
            for topic_idx, topic in enumerate(log.topics):
                if topic in already_checked:
                    continue
                elif uint32.serialize(topic) not in receipt.bloom_filter:
                    raise ValidationError(
                        f"The topic at position {topic_idx} from the log entry at "
                        f"position {log_idx} is not present in the provided bloom filter."  # noqa: E501
                    )
                already_checked.add(topic)

    def validate_block(self, block: BlockAPI) -> None:
        if not isinstance(block, self.get_block_class()):
            raise ValidationError(
                f"This vm ({self!r}) is not equipped to validate a block of type {block!r}"  # noqa: E501
            )

        if block.is_genesis:
            validate_length_lte(
                block.header.extra_data,
                self.extra_data_max_bytes,
                title="BlockHeader.extra_data",
            )
        else:
            parent_header = get_parent_header(block.header, self.chaindb)
            self.validate_header(block.header, parent_header)

        tx_root_hash, _ = make_trie_root_and_nodes(block.transactions)
        if tx_root_hash != block.header.transaction_root:
            raise ValidationError(
                f"Block's transaction_root ({block.header.transaction_root!r}) "
                f"does not match expected value: {tx_root_hash!r}"
            )

        if len(block.uncles) > MAX_UNCLES:
            raise ValidationError(
                f"Blocks may have a maximum of {MAX_UNCLES} uncles.  "
                f"Found {len(block.uncles)}."
            )

        if not self.chaindb.exists(block.header.state_root):
            # If not in the db, check if the current state root matches.
            if not self.state.make_state_root() == block.header.state_root:
                raise ValidationError(
                    "`state_root` does not match or was not found in the db.\n"
                    f"- state_root: {block.header.state_root!r}"
                )

        local_uncle_hash = keccak(rlp.encode(block.uncles))
        if local_uncle_hash != block.header.uncles_hash:
            raise ValidationError(
                "`uncles_hash` and block `uncles` do not match.\n"
                f" - num_uncles       : {len(block.uncles)}\n"
                f" - block uncle_hash : {local_uncle_hash!r}\n"
                f" - header uncle_hash: {block.header.uncles_hash!r}"
            )

    @classmethod
    def validate_header(
        cls, header: BlockHeaderAPI, parent_header: BlockHeaderAPI
    ) -> None:
        if parent_header is None:
            # to validate genesis header, check if it equals canonical header
            # at block number 0
            raise ValidationError(
                "Must have access to parent header to validate current header"
            )
        else:
            validate_length_lte(
                header.extra_data,
                cls.extra_data_max_bytes,
                title="BlockHeader.extra_data",
            )

            cls.validate_gas(header, parent_header)

            if header.block_number != parent_header.block_number + 1:
                raise ValidationError(
                    "Blocks must be numbered consecutively. "
                    f"Block number #{header.block_number} "
                    f"has parent #{parent_header.block_number}"
                )

            # timestamp
            if header.timestamp <= parent_header.timestamp:
                raise ValidationError(
                    "timestamp must be strictly later than parent, "
                    f"but is {parent_header.timestamp - header.timestamp} seconds before.\n"  # noqa: E501
                    f"- child  : {header.timestamp}\n"
                    f"- parent : {parent_header.timestamp}. "
                )

    @classmethod
    def validate_gas(
        cls, header: BlockHeaderAPI, parent_header: BlockHeaderAPI
    ) -> None:
        validate_gas_limit(header.gas_limit, parent_header.gas_limit)

    def validate_seal(self, header: BlockHeaderAPI) -> None:
        try:
            self._consensus.validate_seal(header)
        except ValidationError as exc:
            self.cls_logger.debug(
                "Failed to validate seal on header: %r. Error: %s",
                header.as_dict(),
                exc,
            )
            raise

    def validate_seal_extension(
        self, header: BlockHeaderAPI, parents: Iterable[BlockHeaderAPI]
    ) -> None:
        self._consensus.validate_seal_extension(header, parents)

    @classmethod
    def validate_uncle(
        cls, block: BlockAPI, uncle: BlockHeaderAPI, uncle_parent: BlockHeaderAPI
    ) -> None:
        if uncle.block_number >= block.number:
            raise ValidationError(
                f"Uncle number ({uncle.block_number}) is higher than "
                f"block number ({block.number})"
            )
        if uncle.block_number != uncle_parent.block_number + 1:
            raise ValidationError(
                f"Uncle number ({uncle.block_number}) is not one above "
                f"ancestor's number ({uncle_parent.block_number})"
            )
        if uncle.timestamp <= uncle_parent.timestamp:
            raise ValidationError(
                f"Uncle timestamp ({uncle.timestamp}) is not newer than its "
                f"parent's timestamp ({uncle_parent.timestamp})"
            )
        if uncle.gas_used > uncle.gas_limit:
            raise ValidationError(
                f"Uncle's gas usage ({uncle.gas_used}) is above "
                f"the limit ({uncle.gas_limit})"
            )

        uncle_parent_gas_limit = uncle_parent.gas_limit
        if not hasattr(uncle_parent, "base_fee_per_gas") and hasattr(
            uncle, "base_fee_per_gas"
        ):
            # if Berlin -> London transition, double the parent limit for validation
            uncle_parent_gas_limit *= 2

        validate_gas_limit(uncle.gas_limit, uncle_parent_gas_limit)

    #
    # State
    #
    @classmethod
    def get_state_class(cls) -> Type[StateAPI]:
        if cls._state_class is None:
            raise AttributeError("No `_state_class` has been set for this VM")

        return cls._state_class

    @contextlib.contextmanager
    def in_costless_state(self) -> Iterator[StateAPI]:
        header = self.get_header()

        temp_block = self.generate_block_from_parent_header_and_coinbase(
            header, header.coinbase
        )
        prev_hashes = itertools.chain((header.hash,), self.previous_hashes)

        if hasattr(temp_block.header, "base_fee_per_gas"):
            free_header = temp_block.header.copy(base_fee_per_gas=0)
        else:
            free_header = temp_block.header

        state = self.build_state(
            self.chaindb.db, free_header, self.chain_context, prev_hashes
        )

        snapshot = state.snapshot()
        yield state
        state.revert(snapshot)



from typing import (
    Iterable,
)

from eth.abc import (
    AtomicDatabaseAPI,
    BlockHeaderAPI,
    ConsensusAPI,
)
from eth.typing import (
    Address,
)


class PosConsensus(ConsensusAPI):
    """
    Proof of Stake (PoS) consensus is not achieved in the execution layer (EL).
    This consensus class basically implements the same rules as ``NoProofConsensus``
    but with a more proper distinction.
    """

    def __init__(self, base_db: AtomicDatabaseAPI) -> None:
        pass

    def validate_seal(self, header: BlockHeaderAPI) -> None:
        pass

    def validate_seal_extension(
        self, header: BlockHeaderAPI, parents: Iterable[BlockHeaderAPI]
    ) -> None:
        pass

    @classmethod
    def get_fee_recipient(cls, header: BlockHeaderAPI) -> Address:
        return header.coinbase



import copy
from typing import (
    Dict,
)

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.berlin.opcodes import (
    BERLIN_OPCODES,
)
from eth.vm.forks.byzantium.opcodes import (
    ensure_no_static,
)
from eth.vm.logic import (
    block,
)
from eth.vm.opcode import (
    Opcode,
    as_opcode,
)

from . import (
    storage,
)

UPDATED_OPCODES: Dict[int, Opcode] = {
    opcode_values.BASEFEE: as_opcode(
        gas_cost=constants.GAS_BASE,
        logic_fn=block.basefee,
        mnemonic=mnemonics.BASEFEE,
    ),
    opcode_values.SSTORE: as_opcode(
        gas_cost=constants.GAS_NULL,
        logic_fn=ensure_no_static(storage.sstore_eip3529),
        mnemonic=mnemonics.SSTORE,
    ),
}


LONDON_OPCODES = merge(
    copy.deepcopy(BERLIN_OPCODES),
    UPDATED_OPCODES,
)



from typing import (
    Any,
    Callable,
)

from eth_utils.toolz import (
    curry,
)

from eth._utils.db import (
    get_parent_header,
)
from eth._utils.headers import (
    new_timestamp_from_parent,
)
from eth.abc import (
    BlockHeaderAPI,
    VirtualMachineAPI,
)
from eth.constants import (
    BOMB_EXPONENTIAL_FREE_PERIODS,
    BOMB_EXPONENTIAL_PERIOD,
    DIFFICULTY_ADJUSTMENT_DENOMINATOR,
    DIFFICULTY_MINIMUM,
    EMPTY_UNCLE_HASH,
)
from eth.validation import (
    validate_gt,
    validate_header_params_for_configuration,
)
from eth.vm.forks.frontier.headers import (
    create_frontier_header_from_parent,
)

from .constants import (
    BYZANTIUM_DIFFICULTY_ADJUSTMENT_CUTOFF,
)


@curry
def compute_difficulty(
    bomb_delay: int, parent_header: BlockHeaderAPI, timestamp: int
) -> int:
    """
    https://github.com/ethereum/EIPs/issues/100
    """
    parent_timestamp = parent_header.timestamp
    validate_gt(timestamp, parent_timestamp, title="Header.timestamp")

    parent_difficulty = parent_header.difficulty
    offset = parent_difficulty // DIFFICULTY_ADJUSTMENT_DENOMINATOR

    has_uncles = parent_header.uncles_hash != EMPTY_UNCLE_HASH
    adj_factor = max(
        (
            (2 if has_uncles else 1)
            - ((timestamp - parent_timestamp) // BYZANTIUM_DIFFICULTY_ADJUSTMENT_CUTOFF)
        ),
        -99,
    )
    difficulty = max(parent_difficulty + offset * adj_factor, DIFFICULTY_MINIMUM)
    num_bomb_periods = (
        max(
            0,
            parent_header.block_number + 1 - bomb_delay,
        )
        // BOMB_EXPONENTIAL_PERIOD
    ) - BOMB_EXPONENTIAL_FREE_PERIODS

    if num_bomb_periods >= 0:
        return max(difficulty + 2**num_bomb_periods, DIFFICULTY_MINIMUM)
    else:
        return difficulty


@curry
def create_header_from_parent(
    difficulty_fn: Callable[[BlockHeaderAPI, int], int],
    parent_header: BlockHeaderAPI,
    **header_params: Any,
) -> BlockHeaderAPI:
    if "timestamp" not in header_params:
        header_params["timestamp"] = new_timestamp_from_parent(parent_header)

    if "difficulty" not in header_params:
        header_params["difficulty"] = difficulty_fn(
            parent_header,
            header_params["timestamp"],
        )
    return create_frontier_header_from_parent(parent_header, **header_params)


@curry
def configure_header(
    vm: VirtualMachineAPI,
    difficulty_fn: Callable[[BlockHeaderAPI, int], int] = None,
    **header_params: Any,
) -> BlockHeaderAPI:
    validate_header_params_for_configuration(header_params)

    with vm.get_header().build_changeset(**header_params) as changeset:
        if (
            "timestamp" in header_params
            and changeset.block_number > 0
            # post-POS does not use difficulty_fn
            and difficulty_fn is not None
        ):
            parent_header = get_parent_header(changeset.build_rlp(), vm.chaindb)
            changeset.difficulty = difficulty_fn(
                parent_header,
                header_params["timestamp"],
            )

        header = changeset.commit()
    return header


compute_byzantium_difficulty = compute_difficulty(3000000)
create_byzantium_header_from_parent = create_header_from_parent(
    compute_byzantium_difficulty
)
configure_byzantium_header = configure_header(
    difficulty_fn=compute_byzantium_difficulty
)



from typing import (
    Iterable,
)

from eth_keys import (
    keys,
)
from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    encode_hex,
    to_tuple,
)

from eth._utils.headers import (
    eth_now,
)
from eth.abc import (
    BlockHeaderAPI,
)
from eth.constants import (
    EMPTY_UNCLE_HASH,
    ZERO_ADDRESS,
    ZERO_HASH32,
)

from .constants import (
    ALLOWED_CLIQUE_DIFFICULTIES,
    COMMON_ADDRESS_LENGTH,
    NONCE_AUTH,
    NONCE_DROP,
    SIGNATURE_LENGTH,
    VANITY_LENGTH,
)
from .datatypes import (
    Snapshot,
)


@to_tuple
def get_signers_at_checkpoint(header: BlockHeaderAPI) -> Iterable[Address]:
    """
    Read the list of signers from a checkpoint header.
    """

    signers_length = len(header.extra_data) - VANITY_LENGTH - SIGNATURE_LENGTH

    if signers_length % COMMON_ADDRESS_LENGTH != 0:
        raise ValidationError("Checkpoint header must contain list of signers")

    signer_count = int(
        (len(header.extra_data) - VANITY_LENGTH - SIGNATURE_LENGTH)
        / COMMON_ADDRESS_LENGTH
    )

    for i in range(signer_count):
        yield Address(
            header.extra_data[VANITY_LENGTH + i * COMMON_ADDRESS_LENGTH :][
                :COMMON_ADDRESS_LENGTH
            ]
        )


def get_signature_hash(header: BlockHeaderAPI) -> Hash32:
    """
    Return the hash that is signed by the block producer. It is defined as the hash of
    the ``header`` except that the last 65 bytes of the ``extra_data`` (the signature)
    are removed before calculating the hash.
    """

    if len(header.extra_data) < SIGNATURE_LENGTH:
        raise ValueError("header.extra_data too short to contain signature")

    signature_header: BlockHeaderAPI = header.copy(
        extra_data=header.extra_data[: len(header.extra_data) - SIGNATURE_LENGTH]
    )
    return signature_header.hash


def get_block_signer(header: BlockHeaderAPI) -> Address:
    """
    Return the address of the signer of the ``header``.
    """

    signature_hash = get_signature_hash(header)

    signature_bytes = header.extra_data[-SIGNATURE_LENGTH:]

    signature = keys.Signature(signature_bytes=signature_bytes)

    return signature.recover_public_key_from_msg_hash(
        signature_hash
    ).to_canonical_address()


def is_in_turn(signer: Address, snapshot: Snapshot, header: BlockHeaderAPI) -> bool:
    """
    Return ``True`` if the block was produced *in turn*, otherwise return ``False``.
    """
    sorted_signers = snapshot.get_sorted_signers()

    try:
        offset = sorted_signers.index(signer)
    except ValueError:
        return False
    else:
        return header.block_number % len(sorted_signers) == offset


def sign_block_header(
    header: BlockHeaderAPI, private_key: PrivateKey
) -> BlockHeaderAPI:
    signature_hash = get_signature_hash(header)
    signature = private_key.sign_msg_hash(signature_hash)
    signers = get_signers_at_checkpoint(header)

    signed_extra_data = b"".join(
        (
            header.extra_data[:VANITY_LENGTH],
            b"".join(signers),
            signature.to_bytes(),
        )
    )

    return header.copy(extra_data=signed_extra_data)


def is_checkpoint(block_number: int, epoch_length: int) -> bool:
    """
    Return ``True`` if the given ``block_number`` is a checkpoint, otherwise ``False``.
    """
    return block_number % epoch_length == 0


def validate_header_integrity(header: BlockHeaderAPI, epoch_length: int) -> None:
    if header.timestamp > eth_now():
        raise ValidationError(f"Invalid future timestamp: {header.timestamp}")

    at_checkpoint = is_checkpoint(header.block_number, epoch_length)

    if at_checkpoint and header.coinbase != ZERO_ADDRESS:
        raise ValidationError(
            f"Coinbase must be {encode_hex(ZERO_ADDRESS)} on checkpoints "
            f"but is {encode_hex(header.coinbase)}"
        )

    if header.nonce != NONCE_AUTH and header.nonce != NONCE_DROP:
        raise ValidationError(f"Invalid nonce: {header.nonce!r}")

    if at_checkpoint and header.nonce != NONCE_DROP:
        raise ValidationError(f"Invalid checkpoint nonce: {header.nonce!r}")

    if len(header.extra_data) < VANITY_LENGTH:
        raise ValidationError("Missing vanity bytes in extra data")

    if len(header.extra_data) < VANITY_LENGTH + SIGNATURE_LENGTH:
        raise ValidationError("Missing signature in extra_data")

    signers_length = len(header.extra_data) - VANITY_LENGTH - SIGNATURE_LENGTH

    if not at_checkpoint and signers_length != 0:
        raise ValidationError("Non-checkpoint header must not contain list of signers")

    if at_checkpoint and signers_length % COMMON_ADDRESS_LENGTH != 0:
        raise ValidationError("Checkpoint header must contain list of signers")

    if header.mix_hash != ZERO_HASH32:
        raise ValidationError(f"Invalid mix hash: {header.mix_hash!r}")

    if header.uncles_hash != EMPTY_UNCLE_HASH:
        raise ValidationError(f"Invalid uncle hash: {header.uncles_hash!r}")

    if header.difficulty not in ALLOWED_CLIQUE_DIFFICULTIES:
        raise ValidationError(f"Invalid difficulty: {header.difficulty}")



from typing import Type

from eth.abc import (
    BlockAPI,
    StateAPI,
)

from ..tangerine_whistle import TangerineWhistleVM

from .blocks import SpuriousDragonBlock
from .state import SpuriousDragonState


class SpuriousDragonVM(TangerineWhistleVM):
    # fork name
    fork: str = "spurious-dragon"

    # classes
    block_class: Type[BlockAPI] = SpuriousDragonBlock
    _state_class: Type[StateAPI] = SpuriousDragonState



import copy
from typing import (
    Dict,
)

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.byzantium.opcodes import (
    ensure_no_static,
)
from eth.vm.forks.muir_glacier.opcodes import (
    MUIR_GLACIER_OPCODES,
)
from eth.vm.forks.tangerine_whistle.constants import (
    GAS_SELFDESTRUCT_EIP150,
)
from eth.vm.opcode import (
    Opcode,
    as_opcode,
)

from . import (
    logic,
)

UPDATED_OPCODES: Dict[int, Opcode] = {
    opcode_values.BALANCE: as_opcode(
        gas_cost=constants.GAS_NULL,
        logic_fn=logic.balance_eip2929,
        mnemonic=mnemonics.BALANCE,
    ),
    opcode_values.EXTCODESIZE: as_opcode(
        gas_cost=constants.GAS_NULL,
        logic_fn=logic.extcodesize_eip2929,
        mnemonic=mnemonics.EXTCODESIZE,
    ),
    opcode_values.EXTCODECOPY: as_opcode(
        gas_cost=constants.GAS_NULL,
        logic_fn=logic.extcodecopy_eip2929,
        mnemonic=mnemonics.EXTCODECOPY,
    ),
    opcode_values.EXTCODEHASH: as_opcode(
        gas_cost=constants.GAS_NULL,
        logic_fn=logic.extcodehash_eip2929,
        mnemonic=mnemonics.EXTCODEHASH,
    ),
    opcode_values.SLOAD: as_opcode(
        gas_cost=constants.GAS_NULL,
        logic_fn=logic.sload_eip2929,
        mnemonic=mnemonics.SLOAD,
    ),
    opcode_values.SSTORE: as_opcode(
        logic_fn=ensure_no_static(logic.sstore_eip2929),
        mnemonic=mnemonics.SSTORE,
        gas_cost=constants.GAS_NULL,
    ),
    # System opcodes
    opcode_values.CREATE: logic.CreateEIP2929.configure(
        __name__="opcode:CREATE",
        mnemonic=mnemonics.CREATE,
        gas_cost=constants.GAS_CREATE,
    )(),
    opcode_values.CALL: logic.CallEIP2929.configure(
        __name__="opcode:CALL",
        mnemonic=mnemonics.CALL,
        gas_cost=constants.GAS_NULL,
    )(),
    opcode_values.CALLCODE: logic.CallCodeEIP2929.configure(
        __name__="opcode:CALLCODE",
        mnemonic=mnemonics.CALLCODE,
        gas_cost=constants.GAS_NULL,
    )(),
    opcode_values.DELEGATECALL: logic.DelegateCallEIP2929.configure(
        __name__="opcode:DELEGATECALL",
        mnemonic=mnemonics.DELEGATECALL,
        gas_cost=constants.GAS_NULL,
    )(),
    opcode_values.CREATE2: logic.Create2EIP2929.configure(
        __name__="opcode:CREATE2",
        mnemonic=mnemonics.CREATE2,
        gas_cost=constants.GAS_CREATE,
    )(),
    opcode_values.STATICCALL: logic.StaticCallEIP2929.configure(
        __name__="opcode:STATICCALL",
        mnemonic=mnemonics.STATICCALL,
        gas_cost=constants.GAS_NULL,
    )(),
    opcode_values.SELFDESTRUCT: as_opcode(
        logic_fn=ensure_no_static(logic.selfdestruct_eip2929),
        mnemonic=mnemonics.SELFDESTRUCT,
        gas_cost=GAS_SELFDESTRUCT_EIP150,
    ),
}


BERLIN_OPCODES = merge(
    copy.deepcopy(MUIR_GLACIER_OPCODES),
    UPDATED_OPCODES,
)



from typing import (
    Any,
    Callable,
    List,
)

from eth_utils.curried import (
    apply_formatter_to_array,
    apply_formatters_to_dict,
    apply_formatters_to_sequence,
    encode_hex,
    to_checksum_address,
    to_hex,
)
from eth_utils.toolz import (
    curried,
)

environment_formatter = apply_formatters_to_dict(
    {
        "currentCoinbase": to_checksum_address,
        "previousHash": encode_hex,
        "currentNumber": to_hex,
        "currentDifficulty": to_hex,
        "currentGasLimit": to_hex,
        "currentTimestamp": to_hex,
    }
)


storage_item_formatter: Callable[[List[Any]], List[Any]]
storage_item_formatter = apply_formatters_to_sequence([to_hex, to_hex])
storage_formatter = curried.itemmap(storage_item_formatter)


account_state_formatter = apply_formatters_to_dict(
    {
        "balance": to_hex,
        "nonce": to_hex,
        "code": encode_hex,
        "storage": storage_formatter,
    }
)


state_item_formatter: Callable[[List[Any]], List[Any]]
state_item_formatter = apply_formatters_to_sequence(
    [to_checksum_address, account_state_formatter]
)
state_formatter = curried.itemmap(state_item_formatter)


transaction_group_formatter = apply_formatters_to_dict(
    {
        # all transaction types
        "to": to_checksum_address,
        "data": apply_formatter_to_array(encode_hex),
        "gasLimit": apply_formatter_to_array(to_hex),
        "gasPrice": to_hex,
        "nonce": to_hex,
        "secretKey": encode_hex,
        "value": apply_formatter_to_array(to_hex),
    }
)


execution_formatter = apply_formatters_to_dict(
    {
        "address": to_checksum_address,
        "origin": to_checksum_address,
        "caller": to_checksum_address,
        "code": encode_hex,
        "value": to_hex,
        "data": encode_hex,
        "gasPrice": to_hex,
        "gas": to_hex,
    }
)


expect_element_formatter = apply_formatters_to_dict({"result": state_formatter})
expect_formatter: Callable[[List[Any]], List[Any]]
expect_formatter = apply_formatter_to_array(expect_element_formatter)


test_formatter = apply_formatters_to_dict(
    {
        "env": environment_formatter,
        "pre": state_formatter,
        "transaction": transaction_group_formatter,
        "expect": expect_formatter,
        "exec": execution_formatter,
    }
)
filler_formatter = curried.valmap(test_formatter)


state_post_formatter = apply_formatters_to_dict({"hash": encode_hex})


filled_state_test_formatter = curried.valmap(
    apply_formatters_to_dict(
        {
            "env": environment_formatter,
            "pre": state_formatter,
            "transaction": transaction_group_formatter,
            "post": state_post_formatter,
        }
    )
)

call_create_item_formatter = apply_formatters_to_dict(
    {
        "data": encode_hex,
        "destination": to_checksum_address,
        "gasLimit": to_hex,
        "value": to_hex,
    }
)
call_creates_formatter: Callable[[List[Any]], List[Any]]
call_creates_formatter = apply_formatter_to_array(call_create_item_formatter)

filled_vm_test_formatter = curried.valmap(
    apply_formatters_to_dict(
        {
            "env": environment_formatter,
            "pre": state_formatter,
            "exec": execution_formatter,
            "post": state_formatter,
            "callcreates": call_creates_formatter,
            "logs": encode_hex,
            "gas": to_hex,
            "output": encode_hex,
        }
    )
)



from eth_hash.auto import (
    keccak,
)

from eth import (
    constants,
)
from eth._utils.numeric import (
    ceil32,
)
from eth.abc import (
    ComputationAPI,
)


def sha3(computation: ComputationAPI) -> None:
    start_position, size = computation.stack_pop_ints(2)

    computation.extend_memory(start_position, size)

    sha3_bytes = computation.memory_read_bytes(start_position, size)
    word_count = ceil32(len(sha3_bytes)) // 32

    gas_cost = constants.GAS_SHA3WORD * word_count
    computation.consume_gas(gas_cost, reason="SHA3: word gas cost")

    result = keccak(sha3_bytes)

    computation.stack_push_bytes(result)



from typing import (
    Any,
    Callable,
    Optional,
)

from toolz import (
    curry,
)

from eth.abc import (
    BlockHeaderAPI,
)
from eth.vm.forks.arrow_glacier.headers import (
    create_arrow_glacier_header_from_parent,
)
from eth.vm.forks.byzantium.headers import (
    compute_difficulty,
    configure_header,
)

from .blocks import (
    GrayGlacierBlockHeader,
)

compute_gray_glacier_difficulty = compute_difficulty(11_400_000)
configure_gray_glacier_header = configure_header(
    difficulty_fn=compute_gray_glacier_difficulty
)


@curry
def create_gray_glacier_header_from_parent(
    difficulty_fn: Callable[[BlockHeaderAPI, int], int],
    parent_header: Optional[BlockHeaderAPI],
    **header_params: Any,
) -> BlockHeaderAPI:
    arrow_glacier_validated_header = create_arrow_glacier_header_from_parent(
        difficulty_fn, parent_header, **header_params
    )

    # extract header params validated up to arrow glacier (previous VM) and plug
    # into `GrayGlacierBlockHeader` class
    all_fields = arrow_glacier_validated_header.as_dict()
    return GrayGlacierBlockHeader(**all_fields)



from typing import Optional, Type

from eth_typing import BlockNumber

from eth.abc import (
    BlockAPI,
    StateAPI,
)
from eth.vm.forks.frontier import FrontierVM

from .blocks import HomesteadBlock
from .headers import (
    create_homestead_header_from_parent,
    compute_homestead_difficulty,
    configure_homestead_header,
)
from .state import HomesteadState


class MetaHomesteadVM(FrontierVM):
    support_dao_fork = True
    _dao_fork_block_number: Optional[BlockNumber] = None

    @classmethod
    def get_dao_fork_block_number(cls) -> BlockNumber:
        if cls._dao_fork_block_number is None:
            raise TypeError(
                "HomesteadVM must be configured with a valid `_dao_fork_block_number`"
            )
        return cls._dao_fork_block_number


class HomesteadVM(MetaHomesteadVM):
    # fork name
    fork: str = "homestead"

    # classes
    block_class: Type[BlockAPI] = HomesteadBlock
    _state_class: Type[StateAPI] = HomesteadState

    # method overrides
    create_header_from_parent = staticmethod(create_homestead_header_from_parent)  # type: ignore  # noqa: E501
    compute_difficulty = staticmethod(compute_homestead_difficulty)  # type: ignore
    configure_header = configure_homestead_header



import functools

from eth.abc import (
    ComputationAPI,
)


def swap_XX(computation: ComputationAPI, position: int) -> None:
    """
    Stack item swapping
    """
    computation.stack_swap(position)


swap1 = functools.partial(swap_XX, position=1)
swap2 = functools.partial(swap_XX, position=2)
swap3 = functools.partial(swap_XX, position=3)
swap4 = functools.partial(swap_XX, position=4)
swap5 = functools.partial(swap_XX, position=5)
swap6 = functools.partial(swap_XX, position=6)
swap7 = functools.partial(swap_XX, position=7)
swap8 = functools.partial(swap_XX, position=8)
swap9 = functools.partial(swap_XX, position=9)
swap10 = functools.partial(swap_XX, position=10)
swap11 = functools.partial(swap_XX, position=11)
swap12 = functools.partial(swap_XX, position=12)
swap13 = functools.partial(swap_XX, position=13)
swap14 = functools.partial(swap_XX, position=14)
swap15 = functools.partial(swap_XX, position=15)
swap16 = functools.partial(swap_XX, position=16)



from typing import (
    Type,
)

from eth.rlp.blocks import BaseBlock
from eth.vm.forks.constantinople import (
    ConstantinopleVM,
)
from eth.vm.state import BaseState

from .blocks import IstanbulBlock
from .headers import (
    compute_istanbul_difficulty,
    configure_istanbul_header,
    create_istanbul_header_from_parent,
)
from .state import IstanbulState


class IstanbulVM(ConstantinopleVM):
    # fork name
    fork = "istanbul"

    # classes
    block_class: Type[BaseBlock] = IstanbulBlock
    _state_class: Type[BaseState] = IstanbulState

    # Methods
    create_header_from_parent = staticmethod(create_istanbul_header_from_parent)  # type: ignore  # noqa: E501
    compute_difficulty = staticmethod(compute_istanbul_difficulty)  # type: ignore
    configure_header = configure_istanbul_header



from typing import (
    Any,
    Union,
)

from eth._utils.spoof import (
    SpoofAttributes,
)
from eth.abc import (
    SignedTransactionAPI,
    UnsignedTransactionAPI,
)


class SpoofTransaction(SpoofAttributes):
    def __init__(
        self,
        transaction: Union[SignedTransactionAPI, UnsignedTransactionAPI],
        **overrides: Any
    ) -> None:
        super().__init__(transaction, **overrides)



import binascii
import functools
from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Sequence,
    Tuple,
    Union,
    cast,
)

from eth_typing import (
    Address,
    HexStr,
)
from eth_utils.curried import (
    ValidationError,
    apply_formatter_if,
    apply_formatter_to_array,
    apply_formatters_to_dict,
    big_endian_to_int,
    decode_hex,
    hexstr_if_str,
    is_0x_prefixed,
    is_bytes,
    is_hex,
    is_integer,
    is_string,
    is_text,
    to_bytes,
    to_canonical_address,
    to_dict,
)
from eth_utils.toolz import (
    assoc_in,
    compose,
    concat,
    curried,
    curry,
    identity,
    merge,
)
from mypy_extensions import (
    TypedDict,
)

from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
)
from eth.tools._utils.mappings import (
    deep_merge,
    is_cleanly_mergable,
)
from eth.typing import (
    AccountState,
    GeneralState,
    IntConvertible,
    Normalizer,
    TransactionDict,
    TransactionNormalizer,
)


#
# Primitives
#
@functools.lru_cache(maxsize=1024)
def normalize_int(value: IntConvertible) -> int:
    """
    Robust to integer conversion, handling hex values, string representations,
    and special cases like `0x`.
    """
    if is_integer(value):
        return cast(int, value)
    elif is_bytes(value):
        return big_endian_to_int(cast(bytes, value))
    elif is_hex(value) and is_0x_prefixed(value):  # type: ignore
        # mypy doesn't recognize that is_hex() forces value to be a str
        value = cast(str, value)
        if len(value) == 2:
            return 0
        else:
            return int(value, 16)
    elif is_string(value):
        return int(value)
    else:
        raise TypeError(f"Unsupported type: Got `{type(value)}`")


def normalize_bytes(value: Union[bytes, str]) -> bytes:
    if is_bytes(value):
        return cast(bytes, value)
    elif is_text(value) and is_hex(value):
        return decode_hex(cast(str, value))
    elif is_text(value):
        return b""
    else:
        raise TypeError("Value must be either a string or bytes object")


@functools.lru_cache(maxsize=1024)
def to_int(value: str) -> int:
    """
    Robust to integer conversion, handling hex values, string representations,
    and special cases like `0x`.
    """
    if is_0x_prefixed(value):
        if len(value) == 2:
            return 0
        else:
            return int(value, 16)
    else:
        return int(value)


@functools.lru_cache(maxsize=128)
def normalize_to_address(value: AnyStr) -> Address:
    if value:
        return to_canonical_address(value)
    else:
        return CREATE_CONTRACT_ADDRESS


robust_decode_hex = hexstr_if_str(to_bytes)


#
# Containers
#
def dict_normalizer(
    formatters: Dict[Any, Callable[..., Any]],
    required: Iterable[Any] = None,
    optional: Iterable[Any] = None,
) -> Normalizer:
    all_keys = set(formatters.keys())

    if required is None and optional is None:
        required_set_form = all_keys
    elif required is not None and optional is not None:
        raise ValueError("Both required and optional keys specified")
    elif required is not None:
        required_set_form = set(required)
    elif optional is not None:
        required_set_form = all_keys - set(optional)

    def normalizer(d: Dict[Any, Any]) -> Dict[str, Any]:
        keys = set(d.keys())
        missing_keys = required_set_form - keys
        superfluous_keys = keys - all_keys
        if missing_keys:
            raise KeyError(f"Missing required keys: {', '.join(missing_keys)}")
        if superfluous_keys:
            raise KeyError(f"Superfluous keys: {', '.join(superfluous_keys)}")

        return apply_formatters_to_dict(formatters, d)

    return normalizer


def dict_options_normalizer(normalizers: Iterable[Normalizer]) -> Normalizer:
    def normalize(d: Dict[Any, Any]) -> Dict[str, Any]:
        first_exception = None
        for normalizer in normalizers:
            try:
                normalized = normalizer(d)
            except KeyError as e:
                if not first_exception:
                    first_exception = e
            else:
                return normalized
        assert first_exception is not None
        raise first_exception

    return normalize


#
# Composition
#
def state_definition_to_dict(state_definition: GeneralState) -> AccountState:
    """Convert a state definition to the canonical dict form.

    State can either be defined in the canonical form, or as a list of sub states that
    are then merged to one. Sub states can either be given as dictionaries themselves,
    or as tuples where the last element is the value and all others the keys for this
    value in the nested state dictionary. Example:

    ```
        [
            ("0xaabb", "balance", 3),
            ("0xaabb", "storage", {
                4: 5,
            }),
            "0xbbcc", {
                "balance": 6,
                "nonce": 7
            }
        ]
    ```
    """
    if isinstance(state_definition, Mapping):
        state_dict = state_definition
    elif isinstance(state_definition, Iterable):
        state_dicts = [
            assoc_in({}, state_item[:-1], state_item[-1])
            if not isinstance(state_item, Mapping)
            else state_item
            for state_item in state_definition
        ]
        if not is_cleanly_mergable(*state_dicts):
            raise ValidationError("Some state item is defined multiple times")
        state_dict = deep_merge(*state_dicts)
    else:
        assert TypeError("State definition must either be a mapping or a sequence")

    seen_keys = set(concat(d.keys() for d in state_dict.values()))
    bad_keys = seen_keys - {"balance", "nonce", "storage", "code"}
    if bad_keys:
        raise ValidationError(
            "State definition contains the following invalid "
            f"account fields: {', '.join(bad_keys)}"
        )

    return state_dict


normalize_storage = compose(
    curried.keymap(normalize_int),
    curried.valmap(normalize_int),
)


normalize_state = compose(
    curried.keymap(to_canonical_address),
    curried.valmap(
        dict_normalizer(
            {
                "balance": normalize_int,
                "code": normalize_bytes,
                "nonce": normalize_int,
                "storage": normalize_storage,
            },
            required=[],
        )
    ),
    apply_formatter_if(
        lambda s: isinstance(s, Iterable) and not isinstance(s, Mapping),
        state_definition_to_dict,
    ),
)


normalize_main_transaction = dict_normalizer(
    {
        "data": normalize_bytes,
        "gasLimit": normalize_int,
        "gasPrice": normalize_int,
        "nonce": normalize_int,
        "secretKey": normalize_bytes,
        "to": normalize_to_address,
        "value": normalize_int,
    }
)


normalize_transaction = cast(
    TransactionNormalizer,
    dict_options_normalizer(
        [
            normalize_main_transaction,
        ]
    ),
)


normalize_main_transaction_group = dict_normalizer(
    {
        "data": apply_formatter_to_array(normalize_bytes),
        "gasLimit": apply_formatter_to_array(normalize_int),
        "gasPrice": normalize_int,
        "nonce": normalize_int,
        "secretKey": normalize_bytes,
        "to": normalize_to_address,
        "value": apply_formatter_to_array(normalize_int),
    }
)


normalize_transaction_group = cast(
    TransactionNormalizer,
    dict_options_normalizer(
        [
            normalize_main_transaction_group,
        ]
    ),
)


normalize_execution = dict_normalizer(
    {
        "address": to_canonical_address,
        "origin": to_canonical_address,
        "caller": to_canonical_address,
        "value": normalize_int,
        "data": normalize_bytes,
        "gasPrice": normalize_int,
        "gas": normalize_int,
    }
)


normalize_networks = identity


normalize_call_create_item = dict_normalizer(
    {
        "data": normalize_bytes,
        "destination": to_canonical_address,
        "gasLimit": normalize_int,
        "value": normalize_int,
    }
)
normalize_call_creates: Callable[[Iterable[Any]], Iterable[Any]]
normalize_call_creates = apply_formatter_to_array(normalize_call_create_item)

normalize_log_item = dict_normalizer(
    {
        "address": to_canonical_address,
        "topics": apply_formatter_to_array(normalize_int),
        "data": normalize_bytes,
    }
)
normalize_logs: Callable[[Iterable[Any]], Iterable[Any]]
normalize_logs = apply_formatter_to_array(normalize_log_item)


normalize_main_environment = dict_normalizer(
    {
        "currentCoinbase": to_canonical_address,
        "previousHash": normalize_bytes,
        "currentNumber": normalize_int,
        "currentDifficulty": normalize_int,
        "currentGasLimit": normalize_int,
        "currentTimestamp": normalize_int,
    },
    optional=["previousHash"],
)


normalize_environment = dict_options_normalizer(
    [
        normalize_main_environment,
    ]
)


#
# Fixture Normalizers
#
def normalize_unsigned_transaction(
    transaction: TransactionDict, indexes: Dict[str, Any]
) -> TransactionDict:
    normalized = normalize_transaction_group(transaction)
    return merge(
        normalized,
        {
            # Dynamic key access not yet allowed with TypedDict
            # https://github.com/python/mypy/issues/5359
            transaction_key: normalized[transaction_key][indexes[index_key]]  # type: ignore  # noqa: E501
            for transaction_key, index_key in [
                ("gasLimit", "gas"),
                ("value", "value"),
                ("data", "data"),
            ]
            if index_key in indexes
        },
    )


FixtureAccountDetails = TypedDict(
    "FixtureAccountDetails",
    {
        "balance": HexStr,
        "nonce": HexStr,
        "code": HexStr,
        "storage": Dict[HexStr, HexStr],
    },
)
FixtureAccountState = Dict[Address, FixtureAccountDetails]


def normalize_account_state(account_state: FixtureAccountState) -> AccountState:
    return {
        to_canonical_address(address): {
            "balance": to_int(state["balance"]),
            "code": decode_hex(state["code"]),
            "nonce": to_int(state["nonce"]),
            "storage": {
                to_int(slot): big_endian_to_int(decode_hex(value))
                for slot, value in state["storage"].items()
            },
        }
        for address, state in account_state.items()
    }


def normalize_post_state(postate: FixtureAccountState) -> AccountState:
    # poststate might not be present in some fixtures
    # https://github.com/ethereum/tests/issues/637#issuecomment-534072897
    if postate is None:
        return {}
    else:
        return normalize_account_state(postate)


@to_dict
def normalize_post_state_hash(
    post_state: Dict[str, Any]
) -> Iterable[Tuple[str, bytes]]:
    yield "hash", decode_hex(post_state["hash"])
    if "logs" in post_state:
        yield "logs", decode_hex(post_state["logs"])


@curry
def normalize_statetest_fixture(
    fixture: Dict[str, Any], fork: str, post_state_index: int
) -> Dict[str, Any]:
    post_state = fixture["post"][fork][post_state_index]

    normalized_fixture = {
        "env": normalize_environment(fixture["env"]),
        "pre": normalize_account_state(fixture["pre"]),
        "post": normalize_post_state_hash(post_state),
        "transaction": normalize_unsigned_transaction(
            fixture["transaction"],
            post_state["indexes"],
        ),
    }

    return normalized_fixture


def normalize_exec(exec_params: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "origin": to_canonical_address(exec_params["origin"]),
        "address": to_canonical_address(exec_params["address"]),
        "caller": to_canonical_address(exec_params["caller"]),
        "value": to_int(exec_params["value"]),
        "data": decode_hex(exec_params["data"]),
        "gas": to_int(exec_params["gas"]),
        "gasPrice": to_int(exec_params["gasPrice"]),
    }


def normalize_callcreates(
    callcreates: Sequence[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    return [
        {
            "data": decode_hex(created_call["data"]),
            "destination": (
                to_canonical_address(created_call["destination"])
                if created_call["destination"]
                else CREATE_CONTRACT_ADDRESS
            ),
            "gasLimit": to_int(created_call["gasLimit"]),
            "value": to_int(created_call["value"]),
        }
        for created_call in callcreates
    ]


@to_dict
def normalize_vmtest_fixture(fixture: Dict[str, Any]) -> Iterable[Tuple[str, Any]]:
    yield "env", normalize_environment(fixture["env"])
    yield "exec", normalize_exec(fixture["exec"])
    yield "pre", normalize_account_state(fixture["pre"])

    if "post" in fixture:
        yield "post", normalize_account_state(fixture["post"])

    if "callcreates" in fixture:
        yield "callcreates", normalize_callcreates(fixture["callcreates"])

    if "gas" in fixture:
        yield "gas", to_int(fixture["gas"])

    if "out" in fixture:
        yield "out", decode_hex(fixture["out"])

    if "logs" in fixture:
        yield "logs", decode_hex(fixture["logs"])


def normalize_signed_transaction(transaction: Dict[str, Any]) -> Dict[str, Any]:
    normalized_universal_transaction = {
        "data": robust_decode_hex(transaction["data"]),
        "gasLimit": to_int(transaction["gasLimit"]),
        "nonce": to_int(transaction["nonce"]),
        "r": to_int(transaction["r"]),
        "s": to_int(transaction["s"]),
        "v": to_int(transaction["v"]),
        "to": decode_hex(transaction["to"]),
        "value": to_int(transaction["value"]),
    }
    if "type" in transaction:
        type_id = to_int(transaction["type"])
        if type_id == 1:
            custom_fields = {
                "type": type_id,
                "gasPrice": to_int(transaction["gasPrice"]),
                "chainId": to_int(transaction["chainId"]),
            }
        elif type_id == 2:
            custom_fields = {
                "type": type_id,
                "chainId": to_int(transaction["chainId"]),
                "maxFeePerGas": to_int(transaction["maxFeePerGas"]),
                "maxPriorityFeePerGas": to_int(transaction["maxPriorityFeePerGas"]),
            }
        else:
            raise ValidationError(f"Did not recognize transaction type {type_id}")
    else:
        custom_fields = {
            "gasPrice": to_int(transaction["gasPrice"]),
        }

    return merge(normalized_universal_transaction, custom_fields)


@curry
def normalize_transactiontest_fixture(
    fixture: Dict[str, Any], fork: str
) -> Dict[str, Any]:
    normalized_fixture = {}

    fork_data = fixture["result"][fork]

    try:
        normalized_fixture["txbytes"] = decode_hex(fixture["txbytes"])
    except binascii.Error:
        normalized_fixture["rlpHex"] = fixture["txbytes"]

    if "sender" in fork_data:
        normalized_fixture["sender"] = fork_data["sender"]

    if "hash" in fork_data:
        normalized_fixture["hash"] = fork_data["hash"]

    return normalized_fixture


def normalize_block_header(header: Dict[str, Any]) -> Dict[str, Any]:
    normalized_header = {
        "bloom": big_endian_to_int(decode_hex(header["bloom"])),
        "coinbase": to_canonical_address(header["coinbase"]),
        "difficulty": to_int(header["difficulty"]),
        "extraData": decode_hex(header["extraData"]),
        "gasLimit": to_int(header["gasLimit"]),
        "gasUsed": to_int(header["gasUsed"]),
        "hash": decode_hex(header["hash"]),
        "mixHash": decode_hex(header["mixHash"]),
        "nonce": decode_hex(header["nonce"]),
        "number": to_int(header["number"]),
        "parentHash": decode_hex(header["parentHash"]),
        "receiptTrie": decode_hex(header["receiptTrie"]),
        "stateRoot": decode_hex(header["stateRoot"]),
        "timestamp": to_int(header["timestamp"]),
        "transactionsTrie": decode_hex(header["transactionsTrie"]),
        "uncleHash": decode_hex(header["uncleHash"]),
    }
    if "blocknumber" in header:
        normalized_header["blocknumber"] = to_int(header["blocknumber"])
    if "baseFeePerGas" in header:
        normalized_header["baseFeePerGas"] = to_int(header["baseFeePerGas"])
    if "chainname" in header:
        normalized_header["chainname"] = header["chainname"]
    if "chainnetwork" in header:
        normalized_header["chainnetwork"] = header["chainnetwork"]
    return normalized_header


def normalize_block(block: Dict[str, Any]) -> Dict[str, Any]:
    normalized_block: Dict[str, Any] = {}

    try:
        normalized_block["rlp"] = decode_hex(block["rlp"])
    except ValueError as err:
        normalized_block["rlp_error"] = err

    if "blockHeader" in block:
        normalized_block["blockHeader"] = normalize_block_header(block["blockHeader"])
    if "transactions" in block:
        normalized_block["transactions"] = [
            normalize_signed_transaction(transaction)
            for transaction in block["transactions"]
        ]
    if "expectException" in block:
        normalized_block["expectException"] = block["expectException"]
    return normalized_block


def normalize_blockchain_fixtures(fixture: Dict[str, Any]) -> Dict[str, Any]:
    normalized_fixture = {
        "blocks": [
            normalize_block(block_fixture) for block_fixture in fixture["blocks"]
        ],
        "genesisBlockHeader": normalize_block_header(fixture["genesisBlockHeader"]),
        "lastblockhash": decode_hex(fixture["lastblockhash"]),
        "pre": normalize_account_state(fixture["pre"]),
        "postState": normalize_post_state(fixture.get("postState")),
        "network": fixture["network"],
    }

    if "sealEngine" in fixture:
        normalized_fixture["sealEngine"] = fixture["sealEngine"]

    if "genesisRLP" in fixture:
        normalized_fixture["genesisRLP"] = decode_hex(fixture["genesisRLP"])

    return normalized_fixture



from typing import (
    Any,
    Optional,
)

from eth_utils import (
    ValidationError,
)
from toolz import (
    curry,
)

from eth.abc import (
    BlockHeaderAPI,
)
from eth.constants import (
    POST_MERGE_DIFFICULTY,
    POST_MERGE_MIX_HASH,
    POST_MERGE_NONCE,
)
from eth.vm.forks.byzantium.headers import (
    configure_header,
)
from eth.vm.forks.gray_glacier.headers import (
    compute_gray_glacier_difficulty,
    create_gray_glacier_header_from_parent,
)

from .blocks import (
    ParisBlockHeader,
)


def _validate_and_return_paris_header_param(
    header_param: str,
    actual: Any,
    constant_value: Any,
) -> Any:
    # if a value is passed into `header_params`, validate it's correct; else, set to
    # the defined EIP-3675 constant value for the `header_param`.
    if actual is not None and actual != constant_value:
        raise ValidationError(
            f"Header param '{header_param}' must always be "
            f"{constant_value}, got: {actual}"
        )
    return constant_value


@curry
def create_paris_header_from_parent(
    parent_header: Optional[BlockHeaderAPI],
    **header_params: Any,
) -> BlockHeaderAPI:
    # `mix_hash` is not strictly validated; take the value from the `header_params`,
    # if present; else, set to the EIP-3675-defined constant value.
    header_params["mix_hash"] = header_params.get("mix_hash", POST_MERGE_MIX_HASH)

    # for `difficulty` and `nonce`, if present in `header_params`, validate the value
    # is the expected EIP-3675 value; else, set to the EIP-3675-defined constant value.
    header_params["difficulty"] = _validate_and_return_paris_header_param(
        "difficulty", header_params.get("difficulty"), POST_MERGE_DIFFICULTY
    )
    header_params["nonce"] = _validate_and_return_paris_header_param(
        "nonce", header_params.get("nonce"), POST_MERGE_NONCE
    )

    gray_glacier_validated_header = create_gray_glacier_header_from_parent(
        compute_gray_glacier_difficulty, parent_header, **header_params
    )

    # extract params validated up to gray glacier (previous VM)
    # and plug into a `ParisBlockHeader` class
    all_fields = gray_glacier_validated_header.as_dict()
    return ParisBlockHeader(**all_fields)


configure_paris_header = configure_header()



import enum
from typing import (
    Iterable,
    Tuple,
)

from eth_typing import (
    BlockNumber,
)
from eth_utils import (
    ValidationError,
    to_tuple,
)

from eth.exceptions import (
    GapTrackingCorrupted,
)
from eth.typing import (
    BlockRange,
    ChainGaps,
)


class GapChange(enum.Enum):
    NoChange = enum.auto()
    NewGap = enum.auto()
    GapFill = enum.auto()
    GapSplit = enum.auto()
    GapLeftShrink = enum.auto()
    GapRightShrink = enum.auto()
    TailWrite = enum.auto()


GAP_WRITES = (
    GapChange.GapFill,
    GapChange.GapSplit,
    GapChange.GapLeftShrink,
    GapChange.GapRightShrink,
)
GENESIS_CHAIN_GAPS = ((), BlockNumber(1))

GapInfo = Tuple[GapChange, ChainGaps]


@to_tuple
def _join_overlapping_gaps(
    unjoined_gaps: Tuple[BlockRange, ...]
) -> Iterable[BlockRange]:
    """
    After introducing a new gap, join any that overlap.
    Input must already be sorted.
    """
    unyielded_low = None
    unyielded_high = None
    for low, high in unjoined_gaps:
        if unyielded_high is not None:
            if low < unyielded_low:
                raise ValidationError(f"Unsorted input! {unjoined_gaps!r}")
            elif unyielded_low <= low <= unyielded_high + 1:
                unyielded_high = max(high, unyielded_high)
                continue
            else:
                yield unyielded_low, unyielded_high

        unyielded_low = low
        unyielded_high = high

    if unyielded_high is not None:
        yield unyielded_low, unyielded_high


def reopen_gap(decanonicalized: BlockNumber, base_gaps: ChainGaps) -> ChainGaps:
    """
    Add a new gap, for a header that was decanonicalized.
    """
    current_gaps, tip_child = base_gaps

    if tip_child <= decanonicalized:
        return base_gaps

    new_raw_gaps = current_gaps + ((decanonicalized, decanonicalized),)

    # join overlapping gaps
    joined_gaps = _join_overlapping_gaps(sorted(new_raw_gaps))

    # is the last gap overlapping with the tip child? if so, merge it
    if joined_gaps[-1][1] + 1 >= tip_child:
        return joined_gaps[:-1], joined_gaps[-1][0]
    else:
        return joined_gaps, tip_child


def is_block_number_in_gap(block_number: BlockNumber, gaps: ChainGaps) -> bool:
    """
    Check if a block number is found in the given gaps
    """
    gap_ranges, tip_child = gaps
    for low, high in gap_ranges:
        if low > block_number:
            return False
        elif high >= block_number:
            return True
        # this range was below the block number, continue looking at the next range

    return block_number >= tip_child


def fill_gap(newly_persisted: BlockNumber, base_gaps: ChainGaps) -> GapInfo:
    """
    Remove a gap, for a new header that was canonicalized.
    """

    current_gaps, tip_child = base_gaps

    if newly_persisted == tip_child:
        # This is adding a consecutive header at the very tail
        new_gaps = (current_gaps, BlockNumber(newly_persisted + 1))
        gap_change = GapChange.TailWrite
    elif newly_persisted > tip_child:
        # We are creating a gap in the chain
        gap_end = BlockNumber(newly_persisted - 1)
        new_gaps = (
            current_gaps + ((tip_child, gap_end),),
            BlockNumber(newly_persisted + 1),
        )
        gap_change = GapChange.NewGap
    elif newly_persisted < tip_child:
        # We are patching a gap which may either shrink an existing gap or divide it
        matching_gaps = [
            (index, pair)
            for index, pair in enumerate(current_gaps)
            if newly_persisted >= pair[0] and newly_persisted <= pair[1]
        ]

        if len(matching_gaps) > 1:
            first_match, second_match, *_ = matching_gaps
            raise GapTrackingCorrupted(
                "Corrupted chain gap tracking",
                f"No. {newly_persisted} appears to be missing in multiple gaps",
                f"1st gap is {first_match[1]}, 2nd gap is {second_match[1]}",
                f"all matching gaps: {matching_gaps}",
            )
        elif len(matching_gaps) == 0:
            # Looks like we are just overwriting an existing header.
            return GapChange.NoChange, base_gaps
        elif len(matching_gaps) == 1:
            gap_index, gap = matching_gaps[0]
            if newly_persisted == gap[0] and newly_persisted == gap[1]:
                updated_center: Tuple[BlockRange, ...] = ()
                gap_change = GapChange.GapFill
            elif newly_persisted == gap[0]:
                # we are shrinking the gap at the start
                updated_center = (
                    (
                        BlockNumber(gap[0] + 1),
                        gap[1],
                    ),
                )
                gap_change = GapChange.GapLeftShrink
            elif newly_persisted == gap[1]:
                # we are shrinking the gap at the tail
                updated_center = (
                    (
                        gap[0],
                        BlockNumber(gap[1] - 1),
                    ),
                )
                gap_change = GapChange.GapRightShrink
            elif gap[0] < newly_persisted < gap[1]:
                # we are dividing the gap
                first_new_gap = (gap[0], BlockNumber(newly_persisted - 1))
                second_new_gap = (BlockNumber(newly_persisted + 1), gap[1])
                updated_center = (
                    first_new_gap,
                    second_new_gap,
                )
                gap_change = GapChange.GapSplit
            else:
                raise Exception("Invariant")

            before_gap = current_gaps[:gap_index]
            after_gap = current_gaps[gap_index + 1 :]
            new_gaps = (before_gap + updated_center + after_gap, tip_child)

    else:
        raise Exception("Invariant")

    return gap_change, new_gaps



import os
from typing import (
    Callable,
    cast,
)

from eth.abc import (
    SignedTransactionAPI,
    StateAPI,
)
from eth._utils.module_loading import (
    import_string,
)


def get_gas_estimator() -> Callable[[StateAPI, SignedTransactionAPI], int]:
    import_path = os.environ.get(
        "GAS_ESTIMATOR_BACKEND_FUNC",
        "eth.estimators.gas.binary_gas_search_intrinsic_tolerance",
    )
    return cast(
        Callable[[StateAPI, SignedTransactionAPI], int], import_string(import_path)
    )



from eth_keys import (
    keys,
)
from eth_keys.exceptions import (
    BadSignature,
)
from eth_utils import (
    ValidationError,
    big_endian_to_int,
)

from eth import (
    constants,
)
from eth._utils.padding import (
    pad32,
    pad32r,
)
from eth.abc import (
    ComputationAPI,
)
from eth.validation import (
    validate_gte,
    validate_lt_secpk1n,
    validate_lte,
)


def ecrecover(computation: ComputationAPI) -> ComputationAPI:
    computation.consume_gas(constants.GAS_ECRECOVER, reason="ECRecover Precompile")
    data = computation.msg.data_as_bytes
    raw_message_hash = data[:32]
    message_hash = pad32r(raw_message_hash)

    v_bytes = pad32r(data[32:64])
    v = big_endian_to_int(v_bytes)

    r_bytes = pad32r(data[64:96])
    r = big_endian_to_int(r_bytes)

    s_bytes = pad32r(data[96:128])
    s = big_endian_to_int(s_bytes)

    try:
        validate_lt_secpk1n(r, title="ECRecover: R")
        validate_lt_secpk1n(s, title="ECRecover: S")
        validate_lte(v, 28, title="ECRecover: V")
        validate_gte(v, 27, title="ECRecover: V")
    except ValidationError:
        return computation

    canonical_v = v - 27

    try:
        signature = keys.Signature(vrs=(canonical_v, r, s))
        public_key = signature.recover_public_key_from_msg_hash(message_hash)
    except BadSignature:
        return computation

    address = public_key.to_canonical_address()
    padded_address = pad32(address)

    computation.output = padded_address
    return computation



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    NewType,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from eth_typing import (
    Address,
    BlockNumber,
    Hash32,
    HexStr,
)
from mypy_extensions import (
    TypedDict,
)

if TYPE_CHECKING:
    from eth.abc import (  # noqa: F401
        BlockHeaderAPI,
        SignedTransactionAPI,
        VirtualMachineAPI,
        WithdrawalAPI,
    )


JournalDBCheckpoint = NewType("JournalDBCheckpoint", int)

AccountDetails = TypedDict(
    "AccountDetails",
    {"balance": int, "nonce": int, "code": bytes, "storage": Dict[int, int]},
)
AccountState = Dict[Address, AccountDetails]
AccountDiff = Iterable[Tuple[Address, str, Union[int, bytes], Union[int, bytes]]]


class Block(TypedDict, total=False):
    header: "BlockHeaderAPI"
    transactions: Sequence["SignedTransactionAPI"]
    uncles: Sequence["BlockHeaderAPI"]
    withdrawals: Sequence["WithdrawalAPI"]


BlockRange = Tuple[BlockNumber, BlockNumber]

ChainGaps = Tuple[
    # series of gaps before the chain head
    Tuple[BlockRange, ...],
    # the first missing block number at the tip of the chain
    BlockNumber,
]

GeneralState = Union[
    AccountState, List[Tuple[Address, Dict[str, Union[int, bytes, Dict[int, int]]]]]
]

GenesisDict = Dict[str, Union[int, BlockNumber, bytes, Hash32]]

BytesOrView = Union[bytes, memoryview]

Normalizer = Callable[[Dict[Any, Any]], Dict[str, Any]]

RawAccountDetails = TypedDict(
    "RawAccountDetails",
    {
        "balance": HexStr,
        "nonce": HexStr,
        "code": HexStr,
        "storage": Dict[HexStr, HexStr],
    },
)

TransactionDict = TypedDict(
    "TransactionDict",
    {
        "nonce": int,
        "gasLimit": int,
        "gasPrice": int,
        "to": Address,
        "value": int,
        "data": bytes,
        "secretKey": bytes,
    },
)

TransactionNormalizer = Callable[[TransactionDict], TransactionDict]

VMFork = Tuple[BlockNumber, Type["VirtualMachineAPI"]]

VMConfiguration = Sequence[VMFork]

VRS = NewType("VRS", Tuple[int, int, int])

IntConvertible = Union[int, bytes, HexStr, str]


TFunc = TypeVar("TFunc")


class StaticMethod(Generic[TFunc]):
    """
    A property class purely to convince mypy to let us assign a function to an
    instance variable.
    See more at: https://github.com/python/mypy/issues/708#issuecomment-405812141
    """

    def __get__(self, oself: Any, owner: Any) -> TFunc:
        return self._func

    def __set__(self, oself: Any, value: TFunc) -> None:
        self._func = value


HeaderParams = Union[Optional[int], BlockNumber, bytes, Address, Hash32]



import pytest

from eth.abc import (
    DatabaseAPI,
)


class DatabaseAPITestSuite:
    def test_database_api_get(self, db: DatabaseAPI) -> None:
        db[b"key-1"] = b"value-1"
        assert db.get(b"key-1") == b"value-1"

    def test_database_api_item_getter(self, db: DatabaseAPI) -> None:
        db[b"key-1"] = b"value-1"
        assert db[b"key-1"] == b"value-1"

    def test_database_api_get_missing_key(self, db: DatabaseAPI) -> None:
        assert b"key-1" not in db
        assert db.get(b"key-1") is None

    def test_database_api_item_getter_missing_key(self, db: DatabaseAPI) -> None:
        assert b"key-1" not in db
        with pytest.raises(KeyError):
            db[b"key-1"]

    def test_database_api_set(self, db: DatabaseAPI) -> None:
        db[b"key-1"] = b"value-1"
        assert db[b"key-1"] == b"value-1"
        db[b"key-1"] = b"value-2"
        assert db[b"key-1"] == b"value-2"

    def test_database_api_item_setter(self, db: DatabaseAPI) -> None:
        db.set(b"key-1", b"value-1")
        assert db[b"key-1"] == b"value-1"
        db.set(b"key-1", b"value-2")
        assert db[b"key-1"] == b"value-2"

    def test_database_api_exists(self, db: DatabaseAPI) -> None:
        assert db.exists(b"key-1") is False

        db[b"key-1"] = b"value-1"

        assert db.exists(b"key-1") is True

    def test_database_api_contains_checking(self, db: DatabaseAPI) -> None:
        assert b"key-1" not in db

        db[b"key-1"] = b"value-1"

        assert b"key-1" in db

    def test_database_api_delete(self, db: DatabaseAPI) -> None:
        db[b"key-1"] = b"value-1"

        assert b"key-1" in db

        db.delete(b"key-1")

        assert not db.exists(b"key-1")
        assert b"key-1" not in db

    def test_database_api_item_delete(self, db: DatabaseAPI) -> None:
        db[b"key-1"] = b"value-1"

        assert b"key-1" in db

        del db[b"key-1"]

        assert b"key-1" not in db

    def test_database_api_delete_missing_key(self, db: DatabaseAPI) -> None:
        assert b"key-1" not in db
        db.delete(b"key-1")

    def test_database_api_item_delete_missing_key(self, db: DatabaseAPI) -> None:
        assert b"key-1" not in db
        with pytest.raises(KeyError):
            del db[b"key-1"]



from typing import (
    Type,
)

from eth_typing import (
    Hash32,
)

from eth.abc import (
    StateAPI,
    TransactionExecutorAPI,
)

from ..gray_glacier import (
    GrayGlacierState,
)
from ..gray_glacier.state import (
    GrayGlacierTransactionExecutor,
)
from .computation import (
    ParisComputation,
)


class ParisTransactionExecutor(GrayGlacierTransactionExecutor):
    pass


class ParisState(GrayGlacierState):
    computation_class = ParisComputation
    transaction_executor_class: Type[TransactionExecutorAPI] = ParisTransactionExecutor

    @property
    def mix_hash(self: StateAPI) -> Hash32:
        return self.execution_context.mix_hash



from typing import (
    Tuple,
)

import rlp
from rlp.sedes import (
    CountableList,
    binary,
)

from eth.abc import (
    LogAPI,
)

from .sedes import (
    address,
    uint32,
)


class Log(rlp.Serializable, LogAPI):
    fields = [("address", address), ("topics", CountableList(uint32)), ("data", binary)]

    def __init__(self, address: bytes, topics: Tuple[int, ...], data: bytes) -> None:
        super().__init__(address, topics, data)

    @property
    def bloomables(self) -> Tuple[bytes, ...]:
        return (self.address,) + tuple(uint32.serialize(topic) for topic in self.topics)



import functools
from typing import (
    Any,
    Callable,
    Type,
    TypeVar,
)

from eth_utils import (
    ExtendedDebugLogger,
    get_extended_debug_logger,
)

from eth._utils.datatypes import (
    Configurable,
)
from eth.abc import (
    ComputationAPI,
    OpcodeAPI,
)

T = TypeVar("T")


class Opcode(Configurable, OpcodeAPI):
    mnemonic: str = None
    gas_cost: int = None

    def __init__(self) -> None:
        if self.mnemonic is None:
            raise TypeError(f"Opcode class {type(self)} missing opcode mnemonic")
        if self.gas_cost is None:
            raise TypeError(f"Opcode class {type(self)} missing opcode gas_cost")

    @property
    def logger(self) -> ExtendedDebugLogger:
        return get_extended_debug_logger(f"eth.vm.logic.{self.mnemonic}")

    @classmethod
    def as_opcode(
        cls: Type[T], logic_fn: Callable[..., Any], mnemonic: str, gas_cost: int
    ) -> T:
        if gas_cost:

            @functools.wraps(logic_fn)
            def wrapped_logic_fn(computation: ComputationAPI) -> Any:
                """
                Wrapper function for the logic function which consumes the base
                opcode gas cost prior to execution.
                """
                computation.consume_gas(
                    gas_cost,
                    mnemonic,
                )
                return logic_fn(computation)

        else:
            wrapped_logic_fn = logic_fn

        props = {
            "__call__": staticmethod(wrapped_logic_fn),
            "mnemonic": mnemonic,
            "gas_cost": gas_cost,
        }
        opcode_cls = type(f"opcode:{mnemonic}", (cls,), props)
        return opcode_cls()

    def __copy__(self) -> "Opcode":
        return type(self)()

    def __deepcopy__(self, memo: Any) -> "Opcode":
        return type(self)()


as_opcode = Opcode.as_opcode



import logging
from typing import (
    Iterable,
    Sequence,
)

from eth_typing import (
    Address,
)
from eth_utils import (
    ValidationError,
    encode_hex,
    to_tuple,
)

from eth.abc import (
    AtomicDatabaseAPI,
    BlockHeaderAPI,
    ConsensusAPI,
    ConsensusContextAPI,
    VirtualMachineAPI,
    VirtualMachineModifierAPI,
)
from eth.db.chain import (
    ChainDB,
)
from eth.typing import (
    HeaderParams,
    VMConfiguration,
    VMFork,
)

from ._utils import (
    get_block_signer,
    is_in_turn,
    validate_header_integrity,
)
from .constants import (
    EPOCH_LENGTH,
)
from .datatypes import (
    Snapshot,
)
from .snapshot_manager import (
    SnapshotManager,
)


def configure_header(
    vm: VirtualMachineAPI, **header_params: HeaderParams
) -> BlockHeaderAPI:
    with vm.get_header().build_changeset(**header_params) as changeset:
        # We do this because the default `configure_header` tries to compute the
        # difficulty which we can not do at this point. We take the `difficulty` as
        # provided and rely on the `validate_seal` call that will happen as the last
        # step when blocks are imported.
        header = changeset.commit()
    return header


def _construct_turn_error_message(
    expected_difficulty: int,
    header: BlockHeaderAPI,
    signer: Address,
    signers: Sequence[Address],
) -> str:
    return (
        f"Expected difficulty of {header} to be {expected_difficulty} "
        f"but was {header.difficulty}.\n"
        f"Header signer: {encode_hex(signer)}. Valid signers: {signers}"
    )


class CliqueConsensusContext(ConsensusContextAPI):
    epoch_length = EPOCH_LENGTH

    def __init__(self, db: AtomicDatabaseAPI):
        self.db = db
        self.snapshot_manager = SnapshotManager(ChainDB(db), self.epoch_length)


class CliqueConsensus(ConsensusAPI):
    """
    This class is the entry point to operate a chain under the rules of Clique consensus
    which is defined in EIP-225: https://eips.ethereum.org/EIPS/eip-225
    """

    logger = logging.getLogger("eth.consensus.clique.CliqueConsensus")

    def __init__(self, context: CliqueConsensusContext) -> None:
        if context is None:
            raise ValueError("Can not instantiate without `context`")
        self._epoch_length = context.epoch_length
        self._snapshot_manager = context.snapshot_manager

    @classmethod
    def get_fee_recipient(cls, header: BlockHeaderAPI) -> Address:
        """
        If the ``header`` has a signer, return the signer, otherwise return the
        ``coinbase`` of the passed header.
        """
        try:
            return get_block_signer(header)
        except ValueError:
            return header.coinbase

    def get_snapshot(self, header: BlockHeaderAPI) -> Snapshot:
        """
        Retrieve the ``Snapshot`` for the given ``header``.
        """
        return self._snapshot_manager.get_or_create_snapshot(
            header.block_number, header.hash
        )

    def validate_seal(self, header: BlockHeaderAPI) -> None:
        """
        Only validate the integrity of the header, use `validate_seal_extension`
        to validate the consensus relevant seal of the header.
        """
        validate_header_integrity(header, self._epoch_length)

    def validate_seal_extension(
        self, header: BlockHeaderAPI, parents: Iterable[BlockHeaderAPI]
    ) -> None:
        """
        Validate the seal of the given ``header`` according
        to the Clique consensus rules.
        """

        if header.block_number == 0:
            return

        validate_header_integrity(header, self._epoch_length)

        signer = get_block_signer(header)
        snapshot = self._snapshot_manager.get_or_create_snapshot(
            header.block_number - 1, header.parent_hash, parents
        )
        in_turn = is_in_turn(signer, snapshot, header)

        authorized_signers = snapshot.get_sorted_signers()

        if in_turn and header.difficulty != 2:
            raise ValidationError(
                _construct_turn_error_message(2, header, signer, authorized_signers)
            )
        elif not in_turn and header.difficulty != 1:
            raise ValidationError(
                _construct_turn_error_message(1, header, signer, authorized_signers)
            )

        self._snapshot_manager.apply(snapshot, header)

        if signer not in authorized_signers:
            raise ValidationError(
                f"Failed to validate {header}."
                f"Signer {encode_hex(signer)} not in {authorized_signers}"
            )


class CliqueApplier(VirtualMachineModifierAPI):
    """
    This class is used to apply a clique consensus engine
    to a series of virtual machines
    """

    @to_tuple
    def amend_vm_configuration(self, config: VMConfiguration) -> Iterable[VMFork]:
        """
        Amend the given ``VMConfiguration`` to operate
        under the rules of Clique consensus.
        """
        for pair in config:
            block_number, vm = pair
            vm_class = vm.configure(
                extra_data_max_bytes=65535,
                consensus_class=CliqueConsensus,
                configure_header=configure_header,
                get_block_reward=staticmethod(int),
                get_uncle_reward=staticmethod(int),
            )

            yield block_number, vm_class



import contextlib
from typing import (
    Iterator,
    cast,
)

from eth_hash.auto import (
    keccak,
)
from trie import (
    HexaryTrie,
)

from eth.db.keymap import (
    KeyMapDB,
)


class HashTrie(KeyMapDB):
    keymap = keccak  # type: ignore  # mypy doesn't like that keccak accepts bytearray

    @contextlib.contextmanager
    def squash_changes(self) -> Iterator["HashTrie"]:
        with cast(HexaryTrie, self._db).squash_changes() as memory_trie:
            yield type(self)(memory_trie)



import functools

from eth.abc import (
    ComputationAPI,
)


def pop(computation: ComputationAPI) -> None:
    computation.stack_pop1_any()


def push_XX(computation: ComputationAPI, size: int) -> None:
    raw_value = computation.code.read(size)

    # This is a performance-sensitive area.
    # Calling raw_value.ljust() when size == len(raw_value) is more expensive than
    # calling len(raw_value) and raw_len is typically the correct size already,
    # so this saves a bit of time:
    raw_len = len(raw_value)
    if raw_len == size:
        computation.stack_push_bytes(raw_value)
    else:
        padded_value = raw_value.ljust(size, b"\x00")
        computation.stack_push_bytes(padded_value)


push0 = functools.partial(push_XX, size=0)
push1 = functools.partial(push_XX, size=1)
push2 = functools.partial(push_XX, size=2)
push3 = functools.partial(push_XX, size=3)
push4 = functools.partial(push_XX, size=4)
push5 = functools.partial(push_XX, size=5)
push6 = functools.partial(push_XX, size=6)
push7 = functools.partial(push_XX, size=7)
push8 = functools.partial(push_XX, size=8)
push9 = functools.partial(push_XX, size=9)
push10 = functools.partial(push_XX, size=10)
push11 = functools.partial(push_XX, size=11)
push12 = functools.partial(push_XX, size=12)
push13 = functools.partial(push_XX, size=13)
push14 = functools.partial(push_XX, size=14)
push15 = functools.partial(push_XX, size=15)
push16 = functools.partial(push_XX, size=16)
push17 = functools.partial(push_XX, size=17)
push18 = functools.partial(push_XX, size=18)
push19 = functools.partial(push_XX, size=19)
push20 = functools.partial(push_XX, size=20)
push21 = functools.partial(push_XX, size=21)
push22 = functools.partial(push_XX, size=22)
push23 = functools.partial(push_XX, size=23)
push24 = functools.partial(push_XX, size=24)
push25 = functools.partial(push_XX, size=25)
push26 = functools.partial(push_XX, size=26)
push27 = functools.partial(push_XX, size=27)
push28 = functools.partial(push_XX, size=28)
push29 = functools.partial(push_XX, size=29)
push30 = functools.partial(push_XX, size=30)
push31 = functools.partial(push_XX, size=31)
push32 = functools.partial(push_XX, size=32)



from typing import (
    List,
    Optional,
)

from eth_typing import (
    Address,
    BlockNumber,
    Hash32,
)

from eth._warnings import (
    catch_and_ignore_import_warning,
)

with catch_and_ignore_import_warning():
    from eth_utils import (
        denoms,
    )


ANY = "any"
UINT256 = "uint256"
BYTES = "bytes"

UINT_256_MAX = 2**256 - 1
UINT_256_CEILING = 2**256
UINT_255_MAX = 2**255 - 1
UINT_255_CEILING = 2**255
UINT_255_NEGATIVE_ONE = -1 + UINT_256_CEILING
UINT_64_MAX = 2**64 - 1
NULL_BYTE = b"\x00"
EMPTY_WORD = NULL_BYTE * 32

UINT_160_CEILING = 2**160

CREATE_CONTRACT_ADDRESS = Address(b"")
ZERO_ADDRESS = Address(20 * b"\x00")
ZERO_HASH32 = Hash32(32 * b"\x00")


#
# Stack Limit
#
STACK_DEPTH_LIMIT = 1024


#
# Gas Costs and Refunds
#
GAS_NULL = 0
GAS_ZERO = 0
GAS_BASE = 2
GAS_VERYLOW = 3
GAS_LOW = 5
GAS_MID = 8
GAS_HIGH = 10
GAS_EXTCODE = 20
GAS_BALANCE = 20
GAS_SLOAD = 50
GAS_JUMPDEST = 1
GAS_SSET = 20000
GAS_SRESET = 5000

REFUND_SCLEAR = 15000

GAS_SELFDESTRUCT = 0
GAS_SELFDESTRUCT_NEWACCOUNT = 25000
GAS_CREATE = 32000
GAS_CALL = 40
GAS_CALLVALUE = 9000
GAS_CALLSTIPEND = 2300
GAS_NEWACCOUNT = 25000
GAS_EXP = 10
GAS_EXPBYTE = 10
GAS_MEMORY = 3
GAS_TXCREATE = 32000
GAS_TXDATAZERO = 4
GAS_TXDATANONZERO = 68
GAS_TX = 21000
GAS_LOG = 375
GAS_LOGDATA = 8
GAS_LOGTOPIC = 375
GAS_SHA3 = 30
GAS_SHA3WORD = 6
GAS_COPY = 3
GAS_BLOCKHASH = 20
GAS_CODEDEPOSIT = 200

GAS_MEMORY_QUADRATIC_DENOMINATOR = 512


#
# Pre-compile contract gas costs
#
GAS_SHA256 = 60
GAS_SHA256WORD = 12

GAS_RIPEMD160 = 600
GAS_RIPEMD160WORD = 120

GAS_IDENTITY = 15
GAS_IDENTITYWORD = 3

GAS_ECRECOVER = 3000

GAS_ECADD = 500
GAS_ECMUL = 40000

GAS_ECPAIRING_BASE = 100000
GAS_ECPAIRING_PER_POINT = 80000


#
# Gas Limit
#
GAS_LIMIT_EMA_DENOMINATOR = 1024
GAS_LIMIT_ADJUSTMENT_FACTOR = 1024
GAS_LIMIT_MINIMUM = 5000
GAS_LIMIT_MAXIMUM = 2**63 - 1

GAS_LIMIT_USAGE_ADJUSTMENT_NUMERATOR = 3
GAS_LIMIT_USAGE_ADJUSTMENT_DENOMINATOR = 2


#
# Difficulty
#
DIFFICULTY_ADJUSTMENT_DENOMINATOR = 2048
DIFFICULTY_MINIMUM = 131072

BOMB_EXPONENTIAL_PERIOD = 100000
BOMB_EXPONENTIAL_FREE_PERIODS = 2

#
# Mining
#
BLOCK_REWARD = 5 * denoms.ether

UNCLE_DEPTH_PENALTY_FACTOR = 8
MAX_UNCLE_DEPTH = 6
MAX_UNCLES = 2

#
# SECPK1N
#
SECPK1_P = 2**256 - 2**32 - 977
SECPK1_N = (
    115792089237316195423570985008687907852837564279074904382605163141518161494337
)
SECPK1_A = 0
SECPK1_B = 7
SECPK1_Gx = (
    55066263022277343669578718895168534326250603453777594175500187360389116729240
)
SECPK1_Gy = (
    32670510020758816978083085130507043184471273380659243275938904335757337482424
)
SECPK1_G = (SECPK1_Gx, SECPK1_Gy)


#
# Block and Header
#
# keccak(rlp.encode([]))
EMPTY_UNCLE_HASH = Hash32(
    b"\x1d\xccM\xe8\xde\xc7]z\xab\x85\xb5g\xb6\xcc\xd4\x1a\xd3\x12E\x1b\x94\x8at\x13\xf0\xa1B\xfd@\xd4\x93G"  # noqa: E501
)


#
# Genesis Data
#
GENESIS_BLOCK_NUMBER = BlockNumber(0)
GENESIS_DIFFICULTY = 17179869184
GENESIS_GAS_LIMIT = 5000
GENESIS_PARENT_HASH = ZERO_HASH32
GENESIS_COINBASE = ZERO_ADDRESS
GENESIS_NONCE = b"\x00\x00\x00\x00\x00\x00\x00B"  # 0x42 encoded as big-endian-integer
GENESIS_MIX_HASH = ZERO_HASH32
GENESIS_EXTRA_DATA = b""
GENESIS_BLOOM = 0
GENESIS_GAS_USED = 0

#
# Sha3 Keccak
#
EMPTY_SHA3 = Hash32(
    b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p"  # noqa: E501
)
BLANK_ROOT_HASH = Hash32(
    b"V\xe8\x1f\x17\x1b\xccU\xa6\xff\x83E\xe6\x92\xc0\xf8n\x5bH\xe0\x1b\x99l\xad\xc0\x01b/\xb5\xe3c\xb4!"  # noqa: E501
)


GAS_MOD_EXP_QUADRATIC_DENOMINATOR = 20

#
# BLOCKHASH opcode maximum depth
#
MAX_PREV_HEADER_DEPTH = 256

#
# Call overrides
#
DEFAULT_SPOOF_Y_PARITY = 1
DEFAULT_SPOOF_R = 1
DEFAULT_SPOOF_S = 1


#
# Merge / EIP-3675 constants
#
POST_MERGE_OMMERS_HASH = EMPTY_UNCLE_HASH
POST_MERGE_DIFFICULTY = 0
POST_MERGE_MIX_HASH = ZERO_HASH32
POST_MERGE_NONCE = b"\x00\x00\x00\x00\x00\x00\x00\x00"
POST_MERGE_OMMERS: List[Optional[Hash32]] = []



from typing import (
    Any,
    Dict,
    Iterable,
    Tuple,
    Union,
)

from eth.tools._utils.hashing import (
    hash_log_entries,
)
from eth.tools._utils.mappings import (
    deep_merge,
)
from eth.tools._utils.normalization import (
    normalize_bytes,
    normalize_call_creates,
    normalize_environment,
    normalize_execution,
    normalize_int,
    normalize_logs,
    normalize_state,
)
from eth.tools.fixtures.helpers import (
    get_test_name,
)


def fill_vm_test(
    filler: Dict[str, Any],
    *,
    call_creates: Any = None,
    gas_price: Union[int, str] = None,
    gas_remaining: Union[int, str] = 0,
    logs: Iterable[Tuple[bytes, Tuple[int, ...], bytes]] = None,
    output: bytes = b""
) -> Dict[str, Dict[str, Any]]:
    test_name = get_test_name(filler)
    test = filler[test_name]

    environment = normalize_environment(test["env"])
    pre_state = normalize_state(test["pre"])
    execution = normalize_execution(test["exec"])

    assert len(test["expect"]) == 1
    expect = test["expect"][0]
    assert "network" not in test
    assert "indexes" not in test

    result = normalize_state(expect["result"])
    post_state = deep_merge(pre_state, result)

    call_creates = normalize_call_creates(call_creates or [])
    gas_remaining = normalize_int(gas_remaining)
    output = normalize_bytes(output)

    logs = normalize_logs(logs or [])
    log_hash = hash_log_entries(logs)

    return {
        test_name: {
            "env": environment,
            "pre": pre_state,
            "exec": execution,
            "post": post_state,
            "callcreates": call_creates,
            "gas": gas_remaining,
            "output": output,
            "logs": log_hash,
        }
    }



from typing import (
    Type,
)

from rlp.sedes import (
    CountableList,
)

from eth.abc import (
    ReceiptBuilderAPI,
    TransactionBuilderAPI,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.vm.forks.muir_glacier.blocks import (
    MuirGlacierBlock,
)

from .receipts import (
    BerlinReceiptBuilder,
)
from .transactions import (
    BerlinTransactionBuilder,
)


class BerlinBlock(MuirGlacierBlock):
    transaction_builder: Type[TransactionBuilderAPI] = BerlinTransactionBuilder  # type: ignore  # noqa: E501
    receipt_builder: Type[ReceiptBuilderAPI] = BerlinReceiptBuilder  # type: ignore
    fields = [
        ("header", BlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(BlockHeader)),
    ]



from .applier import ConsensusApplier  # noqa: F401
from .clique.clique import (  # noqa: F401
    CliqueApplier,
    CliqueConsensus,
    CliqueConsensusContext,
)
from .context import ConsensusContext  # noqa: F401
from .noproof import NoProofConsensus  # noqa: F401
from .pow import PowConsensus  # noqa: F401



from eth_hash.auto import (
    keccak,
)
from eth_utils import (
    encode_hex,
)

from eth import (
    precompiles,
)
from eth._utils.address import (
    force_bytes_to_address,
)
from eth.abc import (
    ComputationAPI,
    MessageAPI,
    StateAPI,
    TransactionContextAPI,
)
from eth.constants import (
    GAS_CODEDEPOSIT,
    STACK_DEPTH_LIMIT,
)
from eth.exceptions import (
    InsufficientFunds,
    OutOfGas,
    StackDepthLimit,
)
from eth.vm.computation import (
    BaseComputation,
)

from .opcodes import (
    FRONTIER_OPCODES,
)

FRONTIER_PRECOMPILES = {
    force_bytes_to_address(b"\x01"): precompiles.ecrecover,
    force_bytes_to_address(b"\x02"): precompiles.sha256,
    force_bytes_to_address(b"\x03"): precompiles.ripemd160,
    force_bytes_to_address(b"\x04"): precompiles.identity,
}


class FrontierComputation(BaseComputation):
    """
    A class for all execution message computations in the ``Frontier`` fork.
    Inherits from :class:`~eth.vm.computation.BaseComputation`
    """

    # Override
    opcodes = FRONTIER_OPCODES
    _precompiles = FRONTIER_PRECOMPILES  # type: ignore # https://github.com/python/mypy/issues/708 # noqa: E501

    @classmethod
    def apply_message(
        cls,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> ComputationAPI:
        snapshot = state.snapshot()

        if message.depth > STACK_DEPTH_LIMIT:
            raise StackDepthLimit("Stack depth limit reached")

        if message.should_transfer_value and message.value:
            sender_balance = state.get_balance(message.sender)

            if sender_balance < message.value:
                raise InsufficientFunds(
                    f"Insufficient funds: {sender_balance} < {message.value}"
                )

            state.delta_balance(message.sender, -1 * message.value)
            state.delta_balance(message.storage_address, message.value)

            cls.logger.debug2(
                "TRANSFERRED: %s from %s -> %s",
                message.value,
                encode_hex(message.sender),
                encode_hex(message.storage_address),
            )

        state.touch_account(message.storage_address)

        computation = cls.apply_computation(
            state,
            message,
            transaction_context,
        )

        if computation.is_error:
            state.revert(snapshot)
        else:
            state.commit(snapshot)

        return computation

    @classmethod
    def apply_create_message(
        cls,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> ComputationAPI:
        computation = cls.apply_message(state, message, transaction_context)

        if computation.is_error:
            return computation
        else:
            contract_code = computation.output

            if contract_code:
                contract_code_gas_fee = len(contract_code) * GAS_CODEDEPOSIT
                try:
                    computation.consume_gas(
                        contract_code_gas_fee,
                        reason="Write contract code for CREATE",
                    )
                except OutOfGas:
                    computation.output = b""
                else:
                    cls.logger.debug2(
                        "SETTING CODE: %s -> length: %s | hash: %s",
                        encode_hex(message.storage_address),
                        len(contract_code),
                        encode_hex(keccak(contract_code)),
                    )
                    state.set_code(message.storage_address, contract_code)
            return computation



from eth_hash.auto import (
    keccak,
)
from eth_utils import (
    encode_hex,
)

from eth import (
    constants,
)
from eth.abc import (
    ComputationAPI,
    MessageAPI,
    StateAPI,
    TransactionContextAPI,
)
from eth.exceptions import (
    OutOfGas,
)
from eth.vm.forks.frontier.computation import (
    FrontierComputation,
)

from .opcodes import (
    HOMESTEAD_OPCODES,
)


class HomesteadComputation(FrontierComputation):
    """
    A class for all execution *message* computations in the ``Frontier`` fork.
    Inherits from :class:`~eth.vm.forks.frontier.computation.FrontierComputation`
    """

    # Override
    opcodes = HOMESTEAD_OPCODES

    @classmethod
    def apply_create_message(
        cls,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> ComputationAPI:
        snapshot = state.snapshot()

        computation = cls.apply_message(state, message, transaction_context)

        if computation.is_error:
            state.revert(snapshot)
            return computation
        else:
            contract_code = computation.output

            if contract_code:
                contract_code_gas_cost = len(contract_code) * constants.GAS_CODEDEPOSIT
                try:
                    computation.consume_gas(
                        contract_code_gas_cost,
                        reason="Write contract code for CREATE",
                    )
                except OutOfGas as err:
                    # Different from Frontier: reverts state on gas failure while
                    # writing contract code.
                    computation.error = err
                    state.revert(snapshot)
                else:
                    if cls.logger:
                        cls.logger.debug2(
                            "SETTING CODE: %s -> length: %s | hash: %s",
                            encode_hex(message.storage_address),
                            len(contract_code),
                            encode_hex(keccak(contract_code)),
                        )

                    state.set_code(message.storage_address, contract_code)
                    state.commit(snapshot)
            else:
                state.commit(snapshot)
            return computation



from eth.vm.forks.petersburg.headers import (
    compute_petersburg_difficulty,
    configure_header,
    create_header_from_parent,
)

compute_istanbul_difficulty = compute_petersburg_difficulty

create_istanbul_header_from_parent = create_header_from_parent(
    compute_istanbul_difficulty
)
configure_istanbul_header = configure_header(difficulty_fn=compute_istanbul_difficulty)



from typing import (
    Type,
)

from eth_hash.auto import (
    keccak,
)
from eth_utils import (
    encode_hex,
)

from eth._utils.address import (
    generate_contract_address,
)
from eth.abc import (
    AccountDatabaseAPI,
    ComputationAPI,
    MessageAPI,
    SignedTransactionAPI,
    TransactionContextAPI,
    TransactionExecutorAPI,
)
from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
)
from eth.db.account import (
    AccountDB,
)
from eth.exceptions import (
    ContractCreationCollision,
)
from eth.vm.message import (
    Message,
)
from eth.vm.state import (
    BaseState,
    BaseTransactionExecutor,
)

from .computation import (
    FrontierComputation,
)
from .constants import (
    MAX_REFUND_QUOTIENT,
    REFUND_SELFDESTRUCT,
)
from .transaction_context import (
    FrontierTransactionContext,
)
from .validation import (
    validate_frontier_transaction,
)


class FrontierTransactionExecutor(BaseTransactionExecutor):
    def validate_transaction(self, transaction: SignedTransactionAPI) -> None:
        # Validate the transaction
        transaction.validate()
        self.vm_state.validate_transaction(transaction)

    def build_evm_message(self, transaction: SignedTransactionAPI) -> MessageAPI:
        # Use vm_state.get_gas_price instead of transaction_context.gas_price so
        #   that we can run get_transaction_result (aka~ eth_call) and estimate_gas.
        #   Both work better if the GASPRICE opcode returns the original real price,
        #   but the sender's balance doesn't actually deduct the gas. This
        #   get_gas_price() will return 0 for eth_call, but
        #   transaction_context.gas_price will return
        #   the same value as the GASPRICE opcode.
        gas_fee = transaction.gas * self.vm_state.get_gas_price(transaction)

        # Buy Gas
        self.vm_state.delta_balance(transaction.sender, -1 * gas_fee)

        # Increment Nonce
        self.vm_state.increment_nonce(transaction.sender)

        # Setup VM Message
        message_gas = transaction.gas - transaction.intrinsic_gas

        if transaction.to == CREATE_CONTRACT_ADDRESS:
            contract_address = generate_contract_address(
                transaction.sender,
                self.vm_state.get_nonce(transaction.sender) - 1,
            )
            data = b""
            code = transaction.data
        else:
            contract_address = None
            data = transaction.data
            code = self.vm_state.get_code(transaction.to)

        self.vm_state.logger.debug2(
            (
                "TRANSACTION: sender: %s | to: %s | value: %s | gas: %s | "
                "gas-price: %s | s: %s | r: %s | y_parity: %s | data-hash: %s"
            ),
            encode_hex(transaction.sender),
            encode_hex(transaction.to),
            transaction.value,
            transaction.gas,
            transaction.gas_price,
            transaction.s,
            transaction.r,
            transaction.y_parity,
            encode_hex(keccak(transaction.data)),
        )

        message = Message(
            gas=message_gas,
            to=transaction.to,
            sender=transaction.sender,
            value=transaction.value,
            data=data,
            code=code,
            create_address=contract_address,
        )
        return message

    def build_computation(
        self, message: MessageAPI, transaction: SignedTransactionAPI
    ) -> ComputationAPI:
        transaction_context = self.vm_state.get_transaction_context(transaction)
        if message.is_create:
            is_collision = self.vm_state.has_code_or_nonce(message.storage_address)

            if is_collision:
                # The address of the newly created contract has *somehow* collided
                # with an existing contract address.
                computation = self.vm_state.get_computation(
                    message, transaction_context
                )
                computation.error = ContractCreationCollision(
                    f"Address collision while creating contract: "
                    f"{encode_hex(message.storage_address)}"
                )
                self.vm_state.logger.debug2(
                    "Address collision while creating contract: %s",
                    encode_hex(message.storage_address),
                )
            else:
                computation = self.vm_state.computation_class.apply_create_message(
                    self.vm_state,
                    message,
                    transaction_context,
                )
        else:
            computation = self.vm_state.computation_class.apply_message(
                self.vm_state,
                message,
                transaction_context,
            )

        return computation

    @classmethod
    def calculate_gas_refund(cls, computation: ComputationAPI, gas_used: int) -> int:
        # Self Destruct Refunds
        num_deletions = len(computation.get_accounts_for_deletion())
        if num_deletions:
            computation.refund_gas(REFUND_SELFDESTRUCT * num_deletions)

        # Gas Refunds
        gas_refunded = computation.get_gas_refund()

        return min(gas_refunded, gas_used // MAX_REFUND_QUOTIENT)

    def finalize_computation(
        self, transaction: SignedTransactionAPI, computation: ComputationAPI
    ) -> ComputationAPI:
        transaction_context = self.vm_state.get_transaction_context(transaction)

        gas_remaining = computation.get_gas_remaining()
        gas_used = transaction.gas - gas_remaining
        gas_refund = self.calculate_gas_refund(computation, gas_used)
        gas_refund_amount = (gas_refund + gas_remaining) * transaction_context.gas_price

        if gas_refund_amount:
            self.vm_state.logger.debug2(
                "TRANSACTION REFUND: %s -> %s",
                gas_refund_amount,
                encode_hex(computation.msg.sender),
            )
            self.vm_state.delta_balance(computation.msg.sender, gas_refund_amount)

        # Beneficiary Fees
        gas_used = transaction.gas - gas_remaining - gas_refund
        transaction_fee = gas_used * self.vm_state.get_tip(transaction)

        # EIP-161:
        # Even if the txn fee is zero, the coinbase is still touched here. Post-merge,
        # with no block reward, in the cases where the txn fee is also zero, the
        # coinbase may end up zeroed after the computation and thus should be marked
        # for deletion since it was touched.
        self.vm_state.logger.debug2(
            "TRANSACTION FEE: %s -> %s",
            transaction_fee,
            encode_hex(self.vm_state.coinbase),
        )
        self.vm_state.delta_balance(self.vm_state.coinbase, transaction_fee)

        # Process Self Destructs
        for account, _ in computation.get_accounts_for_deletion():
            self.vm_state.logger.debug2("DELETING ACCOUNT: %s", encode_hex(account))
            self.vm_state.delete_account(account)

        return computation


class FrontierState(BaseState):
    computation_class: Type[ComputationAPI] = FrontierComputation
    transaction_context_class: Type[TransactionContextAPI] = FrontierTransactionContext
    account_db_class: Type[AccountDatabaseAPI] = AccountDB
    transaction_executor_class: Type[
        TransactionExecutorAPI
    ] = FrontierTransactionExecutor

    def apply_transaction(self, transaction: SignedTransactionAPI) -> ComputationAPI:
        executor = self.get_transaction_executor()
        return executor(transaction)

    def validate_transaction(self, transaction: SignedTransactionAPI) -> None:
        validate_frontier_transaction(self, transaction)



import functools
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Tuple,
)

try:
    from vyper.compile_lll import (
        assembly_to_evm,
        compile_to_assembly,
    )
    from vyper.parser.parser_utils import (
        LLLnode,
    )
except ImportError:
    vyper_available = False
else:
    vyper_available = True


def require_vyper(fn: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(fn)
    def inner(*args: Any, **kwargs: Any) -> Any:
        if vyper_available:
            return fn(*args, **kwargs)
        else:
            raise ImportError("The `{0}` function requires the vyper compiler.")

    return inner


@require_vyper
def compile_vyper_lll(vyper_code: List[Any]) -> Tuple[bytes, Dict[str, Any]]:
    lll_node = LLLnode.from_list(vyper_code)
    assembly = compile_to_assembly(lll_node)
    code = assembly_to_evm(assembly)
    return code



from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    Halt,
    InvalidInstruction,
    InvalidJumpDestination,
)
from eth.vm.opcode_values import (
    JUMPDEST,
)


def stop(computation: ComputationAPI) -> None:
    raise Halt("STOP")


def jump(computation: ComputationAPI) -> None:
    jump_dest = computation.stack_pop1_int()

    computation.code.program_counter = jump_dest

    next_opcode = computation.code.peek()

    if next_opcode != JUMPDEST:
        raise InvalidJumpDestination("Invalid Jump Destination")

    if not computation.code.is_valid_opcode(jump_dest):
        raise InvalidInstruction("Jump resulted in invalid instruction")


def jumpi(computation: ComputationAPI) -> None:
    jump_dest, check_value = computation.stack_pop_ints(2)

    if check_value:
        computation.code.program_counter = jump_dest

        next_opcode = computation.code.peek()

        if next_opcode != JUMPDEST:
            raise InvalidJumpDestination("Invalid Jump Destination")

        if not computation.code.is_valid_opcode(jump_dest):
            raise InvalidInstruction("Jump resulted in invalid instruction")


def jumpdest(computation: ComputationAPI) -> None:
    pass


def program_counter(computation: ComputationAPI) -> None:
    pc = max(computation.code.program_counter - 1, 0)

    computation.stack_push_int(pc)


def gas(computation: ComputationAPI) -> None:
    gas_remaining = computation.get_gas_remaining()

    computation.stack_push_int(gas_remaining)



#
# Arithmetic
#
STOP = "STOP"
ADD = "ADD"
MUL = "MUL"
SUB = "SUB"
DIV = "DIV"
SDIV = "SDIV"
MOD = "MOD"
SMOD = "SMOD"
ADDMOD = "ADDMOD"
MULMOD = "MULMOD"
EXP = "EXP"
SIGNEXTEND = "SIGNEXTEND"
SHL = "SHL"
SHR = "SHR"
SAR = "SAR"
#
# Comparisons
#
LT = "LT"
GT = "GT"
SLT = "SLT"
SGT = "SGT"
EQ = "EQ"
ISZERO = "ISZERO"
AND = "AND"
OR = "OR"
XOR = "XOR"
NOT = "NOT"
BYTE = "BYTE"
#
# Sha3
#
SHA3 = "SHA3"
#
# Environment Information
#
ADDRESS = "ADDRESS"
BALANCE = "BALANCE"
SELFBALANCE = "SELFBALANCE"
ORIGIN = "ORIGIN"
CALLER = "CALLER"
CALLVALUE = "CALLVALUE"
CALLDATALOAD = "CALLDATALOAD"
CALLDATASIZE = "CALLDATASIZE"
CALLDATACOPY = "CALLDATACOPY"
CODESIZE = "CODESIZE"
CODECOPY = "CODECOPY"
GASPRICE = "GASPRICE"
EXTCODESIZE = "EXTCODESIZE"
EXTCODECOPY = "EXTCODECOPY"
EXTCODEHASH = "EXTCODEHASH"
RETURNDATASIZE = "RETURNDATASIZE"
RETURNDATACOPY = "RETURNDATACOPY"
CHAINID = "CHAINID"
#
# Block Information
#
BLOCKHASH = "BLOCKHASH"
COINBASE = "COINBASE"
TIMESTAMP = "TIMESTAMP"
NUMBER = "NUMBER"
DIFFICULTY = "DIFFICULTY"
PREVRANDAO = "PREVRANDAO"
GASLIMIT = "GASLIMIT"
BASEFEE = "BASEFEE"
#
# Stack, Memory, Storage and Flow Operations
#
POP = "POP"
MLOAD = "MLOAD"
MSTORE = "MSTORE"
MSTORE8 = "MSTORE8"
SLOAD = "SLOAD"
SSTORE = "SSTORE"
JUMP = "JUMP"
JUMPI = "JUMPI"
PC = "PC"
MSIZE = "MSIZE"
GAS = "GAS"
JUMPDEST = "JUMPDEST"
REVERT = "REVERT"
#
# Push Operations
#
PUSH0 = "PUSH0"
PUSH1 = "PUSH1"
PUSH2 = "PUSH2"
PUSH3 = "PUSH3"
PUSH4 = "PUSH4"
PUSH5 = "PUSH5"
PUSH6 = "PUSH6"
PUSH7 = "PUSH7"
PUSH8 = "PUSH8"
PUSH9 = "PUSH9"
PUSH10 = "PUSH10"
PUSH11 = "PUSH11"
PUSH12 = "PUSH12"
PUSH13 = "PUSH13"
PUSH14 = "PUSH14"
PUSH15 = "PUSH15"
PUSH16 = "PUSH16"
PUSH17 = "PUSH17"
PUSH18 = "PUSH18"
PUSH19 = "PUSH19"
PUSH20 = "PUSH20"
PUSH21 = "PUSH21"
PUSH22 = "PUSH22"
PUSH23 = "PUSH23"
PUSH24 = "PUSH24"
PUSH25 = "PUSH25"
PUSH26 = "PUSH26"
PUSH27 = "PUSH27"
PUSH28 = "PUSH28"
PUSH29 = "PUSH29"
PUSH30 = "PUSH30"
PUSH31 = "PUSH31"
PUSH32 = "PUSH32"
#
# Duplicate Operations
#
DUP1 = "DUP1"
DUP2 = "DUP2"
DUP3 = "DUP3"
DUP4 = "DUP4"
DUP5 = "DUP5"
DUP6 = "DUP6"
DUP7 = "DUP7"
DUP8 = "DUP8"
DUP9 = "DUP9"
DUP10 = "DUP10"
DUP11 = "DUP11"
DUP12 = "DUP12"
DUP13 = "DUP13"
DUP14 = "DUP14"
DUP15 = "DUP15"
DUP16 = "DUP16"
#
# Exchange Operations
#
SWAP1 = "SWAP1"
SWAP2 = "SWAP2"
SWAP3 = "SWAP3"
SWAP4 = "SWAP4"
SWAP5 = "SWAP5"
SWAP6 = "SWAP6"
SWAP7 = "SWAP7"
SWAP8 = "SWAP8"
SWAP9 = "SWAP9"
SWAP10 = "SWAP10"
SWAP11 = "SWAP11"
SWAP12 = "SWAP12"
SWAP13 = "SWAP13"
SWAP14 = "SWAP14"
SWAP15 = "SWAP15"
SWAP16 = "SWAP16"
#
# Logging
#
LOG0 = "LOG0"
LOG1 = "LOG1"
LOG2 = "LOG2"
LOG3 = "LOG3"
LOG4 = "LOG4"
#
# System
#
CREATE = "CREATE"
CREATE2 = "CREATE2"
CALL = "CALL"
CALLCODE = "CALLCODE"
STATICCALL = "STATICCALL"
RETURN = "RETURN"
DELEGATECALL = "DELEGATECALL"
SELFDESTRUCT = "SELFDESTRUCT"



import itertools
from typing import (
    Iterable,
)

from eth_bloom import (
    BloomFilter,
)
import rlp
from rlp.sedes import (
    CountableList,
    big_endian_int,
    binary,
)

from eth.abc import (
    ReceiptAPI,
    ReceiptBuilderAPI,
)

from .logs import (
    Log,
)
from .sedes import (
    uint256,
)


class Receipt(rlp.Serializable, ReceiptAPI, ReceiptBuilderAPI):
    type_id = None

    fields = [
        ("state_root", binary),
        ("gas_used", big_endian_int),
        ("bloom", uint256),
        ("logs", CountableList(Log)),
    ]

    def __init__(
        self, state_root: bytes, gas_used: int, logs: Iterable[Log], bloom: int = None
    ) -> None:
        if bloom is None:
            bloomables = itertools.chain.from_iterable(log.bloomables for log in logs)
            bloom = int(BloomFilter.from_iterable(bloomables))

        super().__init__(
            state_root=state_root,
            gas_used=gas_used,
            bloom=bloom,
            logs=logs,
        )

    @property
    def bloom_filter(self) -> BloomFilter:
        return BloomFilter(self.bloom)

    @classmethod
    def decode(cls, encoded: bytes) -> ReceiptAPI:
        return rlp.decode(encoded, sedes=cls)

    def encode(self) -> bytes:
        return rlp.encode(self)



from typing import (
    TYPE_CHECKING,
    Any,
)

from eth_typing import (
    Address,
)
from eth_utils import (
    decode_hex,
)

from eth._utils.db import (
    get_parent_header,
)
from eth.abc import (
    BlockHeaderAPI,
)
from eth.constants import (
    BOMB_EXPONENTIAL_FREE_PERIODS,
    BOMB_EXPONENTIAL_PERIOD,
    DIFFICULTY_ADJUSTMENT_DENOMINATOR,
    DIFFICULTY_MINIMUM,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.validation import (
    validate_gt,
    validate_header_params_for_configuration,
)
from eth.vm.forks.frontier.headers import (
    create_frontier_header_from_parent,
)

from .constants import (
    HOMESTEAD_DIFFICULTY_ADJUSTMENT_CUTOFF,
)

if TYPE_CHECKING:
    from eth.vm.forks.homestead import HomesteadVM  # noqa: F401


def compute_homestead_difficulty(parent_header: BlockHeaderAPI, timestamp: int) -> int:
    """
    Computes the difficulty for a homestead block based on the parent block.
    """
    parent_tstamp = parent_header.timestamp
    validate_gt(timestamp, parent_tstamp, title="Header.timestamp")
    offset = parent_header.difficulty // DIFFICULTY_ADJUSTMENT_DENOMINATOR
    sign = max(
        1 - (timestamp - parent_tstamp) // HOMESTEAD_DIFFICULTY_ADJUSTMENT_CUTOFF,
        -99,
    )
    difficulty = max(parent_header.difficulty + offset * sign, DIFFICULTY_MINIMUM)
    num_bomb_periods = (
        (parent_header.block_number + 1) // BOMB_EXPONENTIAL_PERIOD
    ) - BOMB_EXPONENTIAL_FREE_PERIODS

    if num_bomb_periods >= 0:
        return max(difficulty + 2**num_bomb_periods, DIFFICULTY_MINIMUM)
    else:
        return difficulty


def create_homestead_header_from_parent(
    parent_header: BlockHeaderAPI, **header_params: Any
) -> BlockHeader:
    if "difficulty" not in header_params:
        # Use setdefault to ensure the new header has the same timestamp we use to
        # calculate its difficulty.
        header_params.setdefault("timestamp", parent_header.timestamp + 1)
        header_params["difficulty"] = compute_homestead_difficulty(
            parent_header,
            header_params["timestamp"],
        )
    return create_frontier_header_from_parent(parent_header, **header_params)


def configure_homestead_header(vm: "HomesteadVM", **header_params: Any) -> BlockHeader:
    validate_header_params_for_configuration(header_params)

    with vm.get_header().build_changeset(**header_params) as changeset:
        if "timestamp" in header_params and changeset.block_number > 0:
            parent_header = get_parent_header(changeset.build_rlp(), vm.chaindb)
            changeset.difficulty = compute_homestead_difficulty(
                parent_header,
                header_params["timestamp"],
            )

        # In geth the modification of the state in the DAO fork block is performed
        # before any transactions are applied, so doing it here is the closest we
        # get to that. Another alternative would be to do it in Block.mine(), but
        # there we'd need to manually instantiate the State and update
        # header.state_root after we're done.
        if (
            vm.support_dao_fork
            and changeset.block_number == vm.get_dao_fork_block_number()
        ):
            state = vm.state

            for hex_account in dao_drain_list:
                address = Address(decode_hex(hex_account))
                balance = state.get_balance(address)
                state.delta_balance(dao_refund_contract, balance)
                state.set_balance(address, 0)

            # Persist the changes to the database
            state.persist()

            # Update state_root manually
            changeset.state_root = state.state_root

        header = changeset.commit()
    return header


dao_refund_contract = Address(decode_hex("0xbf4ed7b27f1d666546e30d74d50d173d20bca754"))
dao_drain_list = [
    "0xd4fe7bc31cedb7bfb8a345f31e668033056b2728",
    "0xb3fb0e5aba0e20e5c49d252dfd30e102b171a425",
    "0x2c19c7f9ae8b751e37aeb2d93a699722395ae18f",
    "0xecd135fa4f61a655311e86238c92adcd779555d2",
    "0x1975bd06d486162d5dc297798dfc41edd5d160a7",
    "0xa3acf3a1e16b1d7c315e23510fdd7847b48234f6",
    "0x319f70bab6845585f412ec7724b744fec6095c85",
    "0x06706dd3f2c9abf0a21ddcc6941d9b86f0596936",
    "0x5c8536898fbb74fc7445814902fd08422eac56d0",
    "0x6966ab0d485353095148a2155858910e0965b6f9",
    "0x779543a0491a837ca36ce8c635d6154e3c4911a6",
    "0x2a5ed960395e2a49b1c758cef4aa15213cfd874c",
    "0x5c6e67ccd5849c0d29219c4f95f1a7a93b3f5dc5",
    "0x9c50426be05db97f5d64fc54bf89eff947f0a321",
    "0x200450f06520bdd6c527622a273333384d870efb",
    "0xbe8539bfe837b67d1282b2b1d61c3f723966f049",
    "0x6b0c4d41ba9ab8d8cfb5d379c69a612f2ced8ecb",
    "0xf1385fb24aad0cd7432824085e42aff90886fef5",
    "0xd1ac8b1ef1b69ff51d1d401a476e7e612414f091",
    "0x8163e7fb499e90f8544ea62bbf80d21cd26d9efd",
    "0x51e0ddd9998364a2eb38588679f0d2c42653e4a6",
    "0x627a0a960c079c21c34f7612d5d230e01b4ad4c7",
    "0xf0b1aa0eb660754448a7937c022e30aa692fe0c5",
    "0x24c4d950dfd4dd1902bbed3508144a54542bba94",
    "0x9f27daea7aca0aa0446220b98d028715e3bc803d",
    "0xa5dc5acd6a7968a4554d89d65e59b7fd3bff0f90",
    "0xd9aef3a1e38a39c16b31d1ace71bca8ef58d315b",
    "0x63ed5a272de2f6d968408b4acb9024f4cc208ebf",
    "0x6f6704e5a10332af6672e50b3d9754dc460dfa4d",
    "0x77ca7b50b6cd7e2f3fa008e24ab793fd56cb15f6",
    "0x492ea3bb0f3315521c31f273e565b868fc090f17",
    "0x0ff30d6de14a8224aa97b78aea5388d1c51c1f00",
    "0x9ea779f907f0b315b364b0cfc39a0fde5b02a416",
    "0xceaeb481747ca6c540a000c1f3641f8cef161fa7",
    "0xcc34673c6c40e791051898567a1222daf90be287",
    "0x579a80d909f346fbfb1189493f521d7f48d52238",
    "0xe308bd1ac5fda103967359b2712dd89deffb7973",
    "0x4cb31628079fb14e4bc3cd5e30c2f7489b00960c",
    "0xac1ecab32727358dba8962a0f3b261731aad9723",
    "0x4fd6ace747f06ece9c49699c7cabc62d02211f75",
    "0x440c59b325d2997a134c2c7c60a8c61611212bad",
    "0x4486a3d68fac6967006d7a517b889fd3f98c102b",
    "0x9c15b54878ba618f494b38f0ae7443db6af648ba",
    "0x27b137a85656544b1ccb5a0f2e561a5703c6a68f",
    "0x21c7fdb9ed8d291d79ffd82eb2c4356ec0d81241",
    "0x23b75c2f6791eef49c69684db4c6c1f93bf49a50",
    "0x1ca6abd14d30affe533b24d7a21bff4c2d5e1f3b",
    "0xb9637156d330c0d605a791f1c31ba5890582fe1c",
    "0x6131c42fa982e56929107413a9d526fd99405560",
    "0x1591fc0f688c81fbeb17f5426a162a7024d430c2",
    "0x542a9515200d14b68e934e9830d91645a980dd7a",
    "0xc4bbd073882dd2add2424cf47d35213405b01324",
    "0x782495b7b3355efb2833d56ecb34dc22ad7dfcc4",
    "0x58b95c9a9d5d26825e70a82b6adb139d3fd829eb",
    "0x3ba4d81db016dc2890c81f3acec2454bff5aada5",
    "0xb52042c8ca3f8aa246fa79c3feaa3d959347c0ab",
    "0xe4ae1efdfc53b73893af49113d8694a057b9c0d1",
    "0x3c02a7bc0391e86d91b7d144e61c2c01a25a79c5",
    "0x0737a6b837f97f46ebade41b9bc3e1c509c85c53",
    "0x97f43a37f595ab5dd318fb46e7a155eae057317a",
    "0x52c5317c848ba20c7504cb2c8052abd1fde29d03",
    "0x4863226780fe7c0356454236d3b1c8792785748d",
    "0x5d2b2e6fcbe3b11d26b525e085ff818dae332479",
    "0x5f9f3392e9f62f63b8eac0beb55541fc8627f42c",
    "0x057b56736d32b86616a10f619859c6cd6f59092a",
    "0x9aa008f65de0b923a2a4f02012ad034a5e2e2192",
    "0x304a554a310c7e546dfe434669c62820b7d83490",
    "0x914d1b8b43e92723e64fd0a06f5bdb8dd9b10c79",
    "0x4deb0033bb26bc534b197e61d19e0733e5679784",
    "0x07f5c1e1bc2c93e0402f23341973a0e043f7bf8a",
    "0x35a051a0010aba705c9008d7a7eff6fb88f6ea7b",
    "0x4fa802324e929786dbda3b8820dc7834e9134a2a",
    "0x9da397b9e80755301a3b32173283a91c0ef6c87e",
    "0x8d9edb3054ce5c5774a420ac37ebae0ac02343c6",
    "0x0101f3be8ebb4bbd39a2e3b9a3639d4259832fd9",
    "0x5dc28b15dffed94048d73806ce4b7a4612a1d48f",
    "0xbcf899e6c7d9d5a215ab1e3444c86806fa854c76",
    "0x12e626b0eebfe86a56d633b9864e389b45dcb260",
    "0xa2f1ccba9395d7fcb155bba8bc92db9bafaeade7",
    "0xec8e57756626fdc07c63ad2eafbd28d08e7b0ca5",
    "0xd164b088bd9108b60d0ca3751da4bceb207b0782",
    "0x6231b6d0d5e77fe001c2a460bd9584fee60d409b",
    "0x1cba23d343a983e9b5cfd19496b9a9701ada385f",
    "0xa82f360a8d3455c5c41366975bde739c37bfeb8a",
    "0x9fcd2deaff372a39cc679d5c5e4de7bafb0b1339",
    "0x005f5cee7a43331d5a3d3eec71305925a62f34b6",
    "0x0e0da70933f4c7849fc0d203f5d1d43b9ae4532d",
    "0xd131637d5275fd1a68a3200f4ad25c71a2a9522e",
    "0xbc07118b9ac290e4622f5e77a0853539789effbe",
    "0x47e7aa56d6bdf3f36be34619660de61275420af8",
    "0xacd87e28b0c9d1254e868b81cba4cc20d9a32225",
    "0xadf80daec7ba8dcf15392f1ac611fff65d94f880",
    "0x5524c55fb03cf21f549444ccbecb664d0acad706",
    "0x40b803a9abce16f50f36a77ba41180eb90023925",
    "0xfe24cdd8648121a43a7c86d289be4dd2951ed49f",
    "0x17802f43a0137c506ba92291391a8a8f207f487d",
    "0x253488078a4edf4d6f42f113d1e62836a942cf1a",
    "0x86af3e9626fce1957c82e88cbf04ddf3a2ed7915",
    "0xb136707642a4ea12fb4bae820f03d2562ebff487",
    "0xdbe9b615a3ae8709af8b93336ce9b477e4ac0940",
    "0xf14c14075d6c4ed84b86798af0956deef67365b5",
    "0xca544e5c4687d109611d0f8f928b53a25af72448",
    "0xaeeb8ff27288bdabc0fa5ebb731b6f409507516c",
    "0xcbb9d3703e651b0d496cdefb8b92c25aeb2171f7",
    "0x6d87578288b6cb5549d5076a207456a1f6a63dc0",
    "0xb2c6f0dfbb716ac562e2d85d6cb2f8d5ee87603e",
    "0xaccc230e8a6e5be9160b8cdf2864dd2a001c28b6",
    "0x2b3455ec7fedf16e646268bf88846bd7a2319bb2",
    "0x4613f3bca5c44ea06337a9e439fbc6d42e501d0a",
    "0xd343b217de44030afaa275f54d31a9317c7f441e",
    "0x84ef4b2357079cd7a7c69fd7a37cd0609a679106",
    "0xda2fef9e4a3230988ff17df2165440f37e8b1708",
    "0xf4c64518ea10f995918a454158c6b61407ea345c",
    "0x7602b46df5390e432ef1c307d4f2c9ff6d65cc97",
    "0xbb9bc244d798123fde783fcc1c72d3bb8c189413",
    "0x807640a13483f8ac783c557fcdf27be11ea4ac7a",
]



from eth.vm.forks.petersburg.state import (
    PetersburgState,
)

from .computation import (
    IstanbulComputation,
)


class IstanbulState(PetersburgState):
    computation_class = IstanbulComputation



from typing import (
    Type,
)

from eth.rlp.blocks import BaseBlock
from eth.vm.forks import (
    MuirGlacierVM,
)
from eth.vm.state import BaseState

from .blocks import BerlinBlock
from .headers import (
    compute_berlin_difficulty,
    configure_berlin_header,
    create_berlin_header_from_parent,
)
from .state import BerlinState


class BerlinVM(MuirGlacierVM):
    # fork name
    fork = "berlin"

    # classes
    block_class: Type[BaseBlock] = BerlinBlock
    _state_class: Type[BaseState] = BerlinState

    # Methods
    create_header_from_parent = staticmethod(create_berlin_header_from_parent)  # type: ignore  # noqa: E501
    compute_difficulty = staticmethod(compute_berlin_difficulty)  # type: ignore
    configure_header = configure_berlin_header



from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    Address,
)
import rlp

from eth._utils.numeric import (
    int_to_bytes32,
)


def force_bytes_to_address(value: bytes) -> Address:
    trimmed_value = value[-20:]
    padded_value = trimmed_value.rjust(20, b"\x00")
    return Address(padded_value)


def generate_contract_address(address: Address, nonce: int) -> Address:
    return force_bytes_to_address(keccak(rlp.encode([address, nonce])))


def generate_safe_contract_address(
    address: Address, salt: int, call_data: bytes
) -> Address:
    return force_bytes_to_address(
        keccak(b"\xff" + address + int_to_bytes32(salt) + keccak(call_data))
    )



from eth.vm.forks.muir_glacier.headers import (
    compute_muir_glacier_difficulty,
    configure_header,
    create_header_from_parent,
)

compute_berlin_difficulty = compute_muir_glacier_difficulty

create_berlin_header_from_parent = create_header_from_parent(compute_berlin_difficulty)
configure_berlin_header = configure_header(difficulty_fn=compute_berlin_difficulty)
